import pytest
import pyteal as pt
from .mapping import Mapping, MapElement
from beaker.application import Application
from beaker.decorators import external


options = pt.CompileOptions(version=pt.MAX_TEAL_VERSION, mode=pt.Mode.Application)


def test_mapping():
    m = Mapping(pt.abi.Address, pt.abi.Uint64)
    assert m._key_type == pt.abi.Address
    assert m._key_type_spec == pt.abi.AddressTypeSpec()

    assert m._value_type == pt.abi.Uint64
    assert m._value_type_spec == pt.abi.Uint64TypeSpec()

    with pytest.raises(pt.TealTypeError):
        m[pt.abi.String()]

    with pytest.raises(pt.TealTypeError):
        m[pt.Int(1)]

    item = m[pt.Txn.sender()]
    assert isinstance(item, MapElement)

    expected, _ = pt.Seq(
        bx := pt.BoxGet(pt.Txn.sender()), pt.Assert(bx.hasValue()), bx.value()
    ).__teal__(options)
    actual, _ = item.get().__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    v = pt.abi.Uint64()
    expected, _ = v.decode(
        pt.Seq(bx := pt.BoxGet(pt.Txn.sender()), pt.Assert(bx.hasValue()), bx.value())
    ).__teal__(options)
    actual, _ = item.store_into(v).__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    expected, _ = pt.Seq(
        pt.Pop(pt.BoxDelete(pt.Txn.sender())), pt.BoxPut(pt.Txn.sender(), v.encode())
    ).__teal__(options)
    actual, _ = item.set(v).__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealTypeError):
        item.set(pt.abi.String())


def test_app_mapping():
    class T(Application):
        m = Mapping(pt.abi.Address, pt.abi.Uint64)

        @external
        def thing(self, name: pt.abi.Address, *, output: pt.abi.Uint64):
            return self.m[name].store_into(output)

    t = T()
    assert len(t.approval_program) > 0



from typing import cast
from pyteal import CompileOptions, Expr, LeafExpr, Mode, TealBlock, TealOp, TealType, Op

# Credit Julian (RandLabs)


class CustomOp:
    def __init__(self, opcode):
        self.opcode = opcode
        self.mode = Mode.Signature | Mode.Application
        self.min_version = 2

    def __str__(self) -> str:
        return self.opcode


class InlineAssembly(LeafExpr):
    """InlineAssembly can be used to inject TEAL source directly into a PyTEAL program

    Often the generated PyTEAL is not as efficient as it can be. This class offers a way to
    write the most efficient TEAL for code paths that would otherwise be impossible because
    of opcode budget constraints.

    It can also be used to implement methods that are not yet available in the PyTEAL repository

    Args:
        opcode: string containing the teal to inject
        args: any number of PyTEAL expressions to place before this opcode
        type: The type this Expression returns, to help during PyTEAL compilation

    """

    def __init__(
        self, opcode: str, *args: "Expr", type: TealType = TealType.none
    ) -> None:
        super().__init__()
        opcode_with_args = opcode.split(" ")
        self.op = CustomOp(opcode_with_args[0])
        self.type = type
        self.opcode_args = opcode_with_args[1:]
        self.args = args

    def __teal__(self, options: "CompileOptions"):
        op = TealOp(self, cast(Op, self.op), *self.opcode_args)
        return TealBlock.FromOp(options, op, *self.args[::1])

    def __str__(self):
        return "(InlineAssembly: {})".format(self.op.opcode)

    def type_of(self):
        return self.type



from typing import Final
from beaker import (
    Application,
    ApplicationStateValue,
    ReservedApplicationStateValue,
    AccountStateValue,
    ReservedAccountStateValue,
    create,
    opt_in,
    external,
)
from pyteal import abi, TealType, Bytes, Int, Txn

from beaker.state import AccountStateBlob, ApplicationStateBlob


class StateExample(Application):

    declared_app_value: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.bytes,
        default=Bytes(
            "A declared state value that is protected with the `static` flag"
        ),
        descr="A static declared variable, nothing at the protocol level protects it, only the methods defined on ApplicationState do",
        static=True,
    )

    dynamic_app_value: Final[
        ReservedApplicationStateValue
    ] = ReservedApplicationStateValue(
        stack_type=TealType.uint64,
        max_keys=32,
        descr="A dynamic app state variable, with 32 possible keys",
    )

    application_blob: Final[ApplicationStateBlob] = ApplicationStateBlob(
        keys=16,
    )

    declared_account_value: Final[AccountStateValue] = AccountStateValue(
        stack_type=TealType.uint64,
        default=Int(1),
        descr="An int stored for each account that opts in",
    )

    dynamic_account_value: Final[ReservedAccountStateValue] = ReservedAccountStateValue(
        stack_type=TealType.bytes,
        max_keys=8,
        descr="A dynamic state value, allowing 8 keys to be reserved, in this case byte type",
    )

    account_blob: Final[AccountStateBlob] = AccountStateBlob(keys=3)

    @create
    def create(self):
        return self.initialize_application_state()

    @opt_in
    def opt_in(self):
        return self.initialize_account_state()

    @external
    def write_acct_blob(self, v: abi.String):
        return self.account_blob.write(Int(0), v.get())

    @external
    def read_acct_blob(self, *, output: abi.DynamicBytes):
        return output.set(
            self.account_blob.read(Int(0), self.account_blob.blob.max_bytes - Int(1))
        )

    @external
    def write_app_blob(self, v: abi.String):
        return self.application_blob.write(Int(0), v.get())

    @external
    def read_app_blob(self, *, output: abi.DynamicBytes):
        return output.set(
            self.application_blob.read(
                Int(0), self.application_blob.blob.max_bytes - Int(1)
            )
        )

    @external
    def set_app_state_val(self, v: abi.String):
        # This will fail, since it was declared as `static` and initialized to a default value during create
        return self.declared_app_value.set(v.get())

    @external(read_only=True)
    def get_app_state_val(self, *, output: abi.String):
        return output.set(self.declared_app_value)

    @external
    def set_dynamic_app_state_val(self, k: abi.Uint8, v: abi.Uint64):
        # Accessing the key with square brackets, accepts both Expr and an ABI type
        # If the value is an Expr it must evaluate to `TealType.bytes`
        # If the value is an ABI type, the `encode` method is used to convert it to bytes
        return self.dynamic_app_value[k].set(v.get())

    @external(read_only=True)
    def get_dynamic_app_state_val(self, k: abi.Uint8, *, output: abi.Uint64):
        return output.set(self.dynamic_app_value[k])

    @external
    def set_account_state_val(self, v: abi.Uint64):
        # Accessing with `[Txn.sender()]` is redundant but
        # more clear what is happening
        return self.declared_account_value[Txn.sender()].set(v.get())

    @external
    def incr_account_state_val(self, v: abi.Uint64):
        # Omitting [Txn.sender()] just for demo purposes
        return self.declared_account_value.increment(v.get())

    @external(read_only=True)
    def get_account_state_val(self, *, output: abi.Uint64):
        return output.set(self.declared_account_value[Txn.sender()])

    @external
    def set_dynamic_account_state_val(self, k: abi.Uint8, v: abi.String):
        return self.dynamic_account_value[k][Txn.sender()].set(v.get())

    @external(read_only=True)
    def get_dynamic_account_state_val(self, k: abi.Uint8, *, output: abi.String):
        return output.set(self.dynamic_account_value[k][Txn.sender()])


if __name__ == "__main__":
    se = StateExample()
    print(se.approval_program)



import pytest
import pyteal as pt
import beaker as bkr

from beaker.testing import UnitTestingApp, assert_output

from beaker.lib.storage.local_blob import LocalBlob
from beaker.lib.storage.blob import blob_page_size


class LocalBlobTest(UnitTestingApp):
    lb = bkr.ReservedAccountStateValue(pt.TealType.bytes, max_keys=16)
    blob = LocalBlob(keys=[x for x in range(10) if x % 2 == 0])


def test_local_blob_zero():
    class LBZero(LocalBlobTest):
        @bkr.external
        def unit_test(self, *, output: pt.abi.DynamicArray[pt.abi.Byte]):
            return pt.Seq(
                self.blob.zero(),
                (s := pt.abi.String()).set(self.blob.read(pt.Int(0), pt.Int(64))),
                output.decode(s.encode()),
            )

    expected = list(bytes(64))
    assert_output(LBZero(), [], [expected])


def test_local_blob_write_read():
    class LB(LocalBlobTest):
        @bkr.external
        def unit_test(self, *, output: pt.abi.DynamicArray[pt.abi.Byte]):
            return pt.Seq(
                self.blob.zero(),
                self.blob.write(pt.Int(0), pt.Bytes("deadbeef" * 8)),
                (s := pt.abi.String()).set(self.blob.read(pt.Int(32), pt.Int(40))),
                output.decode(s.encode()),
            )

    expected = list(b"deadbeef")
    assert_output(LB(), [], [expected])


def test_local_blob_write_read_boundary():
    class LB(LocalBlobTest):
        @bkr.external
        def unit_test(self, *, output: pt.abi.DynamicArray[pt.abi.Byte]):
            return pt.Seq(
                self.blob.zero(pt.Int(0)),
                self.blob.write(pt.Int(0), pt.BytesZero(pt.Int(blob_page_size * 3))),
                (s := pt.abi.String()).set(self.blob.read(pt.Int(32), pt.Int(40))),
                output.decode(s.encode()),
            )

    expected = list(bytes(8))
    assert_output(LB(), [], [expected])


def test_local_blob_write_read_past_end():
    class LB(LocalBlobTest):
        @bkr.external
        def unit_test(self, *, output: pt.abi.DynamicArray[pt.abi.Byte]):
            return pt.Seq(
                self.blob.zero(),
                self.blob.write(pt.Int(0), pt.Bytes("deadbeef" * 8)),
                (s := pt.abi.String()).set(
                    self.blob.read(pt.Int(0), self.blob.max_bytes)
                ),
                output.decode(s.encode()),
            )

    expected = list(bytes(8))

    with pytest.raises(bkr.client.LogicException):
        assert_output(LB(), [], [expected])


def test_local_blob_set_get():
    num = 123

    class LB(LocalBlobTest):
        @bkr.external
        def unit_test(self, *, output: pt.abi.Uint8):
            return pt.Seq(
                self.blob.zero(),
                self.blob.set_byte(pt.Int(32), pt.Int(num)),
                output.set(self.blob.get_byte(pt.Int(32))),
            )

    expected = [num]
    assert_output(LB(), [], expected)


def test_local_blob_set_past_end():
    num = 123

    class LB(LocalBlobTest):
        @bkr.external
        def unit_test(self, *, output: pt.abi.Uint8):
            return pt.Seq(
                self.blob.zero(),
                self.blob.set_byte(self.blob.max_bytes, pt.Int(num)),
                output.set(self.blob.get_byte(pt.Int(32))),
            )

    expected = [num]

    with pytest.raises(bkr.client.LogicException):
        assert_output(LB(), [], expected)



from typing import Final
from pyteal import (
    If,
    Global,
    TealType,
    abi,
    InnerTxnBuilder,
    Seq,
    Bytes,
    TxnType,
    InnerTxn,
    TxnField,
    Assert,
    Expr,
    Int,
    ScratchVar,
    For,
    Approve,
)

from beaker.application import Application, get_method_signature
from beaker.precompile import AppPrecompile
from beaker.state import ApplicationStateValue
from beaker.consts import Algos
from beaker.decorators import internal, external, Authorize


class TargetApp(Application):
    """Simple app that allows the creator to call `opup` in order to increase its opcode budget"""

    @external(authorize=Authorize.only(Global.creator_address()))
    def opup(self):
        return Approve()


class OpUp(Application):
    """OpUp creates a "target" application to make opup calls against in order to increase our opcode budget."""

    #: The app to be created to receiver opup requests
    target: AppPrecompile = AppPrecompile(TargetApp())

    #: The minimum balance required for this class
    min_balance: Final[Expr] = Algos(0.1)

    #: The id of the app created during `bootstrap`
    opup_app_id: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64, key=Bytes("ouaid"), static=True
    )

    @external
    def opup_bootstrap(self, ptxn: abi.PaymentTransaction, *, output: abi.Uint64):
        """initialize opup with bootstrap to create a target app"""
        return Seq(
            Assert(ptxn.get().amount() >= self.min_balance),
            self.create_opup(),
            output.set(OpUp.opup_app_id),
        )

    @internal(TealType.none)
    def create_opup(self):
        """internal method to create the target application"""
        return Seq(
            InnerTxnBuilder.Begin(),
            InnerTxnBuilder.SetFields(
                {
                    TxnField.type_enum: TxnType.ApplicationCall,
                    TxnField.approval_program: self.target.approval.binary,
                    TxnField.clear_state_program: self.target.clear.binary,
                    TxnField.fee: Int(0),
                }
            ),
            InnerTxnBuilder.Submit(),
            self.opup_app_id.set(InnerTxn.created_application_id()),
        )

    @internal(TealType.none)
    def call_opup(self, n):
        """internal method to issue transactions against the target app"""
        return If(
            n == Int(1),
            self.__call_opup(),
            For(
                (i := ScratchVar()).store(Int(0)),
                i.load() < n,
                i.store(i.load() + Int(1)),
            ).Do(Seq(self.__call_opup())),
        )

    # No decorator, inline it
    def __call_opup(self):
        """internal method to just return the method call to our target app"""
        return InnerTxnBuilder.ExecuteMethodCall(
            app_id=self.opup_app_id,
            method_signature=get_method_signature(TargetApp.opup),
            args=[],
            extra_fields={TxnField.fee: Int(0)},
        )



from typing import Final, cast
from Cryptodome.Hash import keccak
from algosdk.atomic_transaction_composer import (
    AtomicTransactionComposer,
    TransactionWithSigner,
)
from algosdk.future.transaction import *

from pyteal import *
from beaker import *
from beaker.precompile import LSigPrecompile

if __name__ == "__main__":
    from lsig import EthEcdsaVerify, HashValue, Signature
else:
    from .lsig import EthEcdsaVerify, HashValue, Signature


class EthChecker(Application):

    # The lsig that will be responsible for validating the
    # incoming signature against the incoming hash
    # When passed to Precompile, it flags the init of the Application
    # to prevent building approval/clear programs until the precompile is
    # compiled so we have access to compiled information (its address for instance)
    verifier: Final[LSigPrecompile] = LSigPrecompile(EthEcdsaVerify(version=6))

    @external
    def check_eth_sig(
        self, hash: HashValue, signature: Signature, *, output: abi.String
    ):
        return Seq(
            # The precompiled lsig should have its address and binary available
            # here so we can use it to make sure we've been called
            # with the correct lsig
            Assert(Txn.sender() == self.verifier.logic.hash()),
            output.set("lsig validated"),
        )


def demo():
    algod_client = sandbox.get_algod_client()
    acct = sandbox.get_accounts().pop()

    # Create app client
    app_client = client.ApplicationClient(
        algod_client, EthChecker(), signer=acct.signer
    )

    # shouldn't have an approval program yet, since
    # the number of precompiles is > 0
    assert app_client.app.approval_program is None

    # This will first compile the precompiles, then compile the approval program
    # with the precompiles in place
    # not required to call manually since create/update will also do this
    # if necessary
    app_client.build()

    # Now we should have the approval program available
    assert app_client.app.approval_program is not None

    app_client.create()

    # Create a new app client with the lsig signer
    lsig_signer = cast(EthChecker, app_client.app).verifier.signer()
    lsig_client = app_client.prepare(signer=lsig_signer)

    atc = AtomicTransactionComposer()

    # Add a payment just to cover fees
    sp_with_fees = algod_client.suggested_params()
    sp_with_fees.flat_fee = True
    sp_with_fees.fee = consts.milli_algo * 3
    atc.add_transaction(
        TransactionWithSigner(
            txn=PaymentTxn(acct.address, sp_with_fees, acct.address, 0),
            signer=acct.signer,
        )
    )

    message = b"OpenZeppelin"
    m = keccak.new(digest_bits=256)
    m.update(message)
    hash = m.digest()

    sp_no_fee = algod_client.suggested_params()
    sp_no_fee.flat_fee = True
    # V0
    hex_signature = "5d99b6f7f6d1f73d1a26497f2b1c89b24c0993913f86e9a2d02cd69887d9c94f3c880358579d811b21dd1b7fd9bb01c1d81d10e69f0384e675c32b39643be8921b"
    signature = bytes.fromhex(hex_signature)
    atc = lsig_client.add_method_call(
        atc,
        EthChecker.check_eth_sig,
        hash=hash,
        signature=signature,
        suggested_params=sp_no_fee,
    )

    # V1
    hex_signature = "331fe75a821c982f9127538858900d87d3ec1f9f737338ad67cad133fa48feff48e6fa0c18abc62e42820f05943e47af3e9fbe306ce74d64094bdf1691ee53e01c"
    signature = bytes.fromhex(hex_signature)
    atc = lsig_client.add_method_call(
        atc,
        EthChecker.check_eth_sig,
        hash=hash,
        signature=signature,
        suggested_params=sp_no_fee,
    )

    result = atc.execute(algod_client, 4)
    for result in result.abi_results:
        print(result.return_value)


if __name__ == "__main__":
    demo()



from typing import Literal
from pyteal import abi, ScratchVar, Seq, Assert, Int, For, Sha256
from beaker import sandbox
from beaker.decorators import external

if __name__ == "__main__":
    from op_up import OpUp
else:
    from .op_up import OpUp


class ExpensiveApp(OpUp):
    """Do expensive work to demonstrate inheriting from OpUp"""

    @external
    def hash_it(
        self,
        input: abi.String,
        iters: abi.Uint64,
        opup_app: abi.Application = OpUp.opup_app_id,
        *,
        output: abi.StaticBytes[Literal[32]],
    ):
        return Seq(
            Assert(opup_app.application_id() == self.opup_app_id),
            self.call_opup(Int(255)),
            (current := ScratchVar()).store(input.get()),
            For(
                (i := ScratchVar()).store(Int(0)),
                i.load() < iters.get(),
                i.store(i.load() + Int(1)),
            ).Do(current.store(Sha256(current.load()))),
            output.decode(current.load()),
        )


if __name__ == "__main__":
    a, c = ExpensiveApp().compile(sandbox.get_algod_client())
    print(a, c)



from typing import Final
from pyteal import abi, TealType, Global, Approve
from beaker import (
    Application,
    AccountStateValue,
    ApplicationStateValue,
    Authorize,
    create,
    external,
    opt_in,
    close_out,
    delete,
    client,
    sandbox,
    consts,
)
from beaker.client.application_client import ApplicationClient
from beaker.client.logic_error import LogicException


class ClientExample(Application):
    manager: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.bytes, default=Global.creator_address()
    )

    nickname: Final[AccountStateValue] = AccountStateValue(
        stack_type=TealType.bytes, descr="what this user prefers to be called"
    )

    @create
    def create(self):
        return self.initialize_application_state()

    @opt_in
    def opt_in(self):
        # Defaults to sender
        return self.initialize_account_state()

    @close_out
    def close_out(self):
        return Approve()

    @delete(authorize=Authorize.only(manager))
    def delete(self):
        return Approve()

    @external(authorize=Authorize.only(manager))
    def set_manager(self, new_manager: abi.Address):
        return self.manager.set(new_manager.get())

    @external
    def set_nick(self, nick: abi.String):
        return self.nickname.set(nick.get())

    @external(read_only=True)
    def get_nick(self, *, output: abi.String):
        return output.set(self.nickname)


def demo():
    # Set up accounts we'll use
    accts = sandbox.get_accounts()

    acct1 = accts.pop()
    acct2 = accts.pop()

    # Create Application client
    app_client1 = client.ApplicationClient(
        client=sandbox.get_algod_client(), app=ClientExample(), signer=acct1.signer
    )

    # Create the app on-chain (uses signer1)
    app_client1.create()
    print(f"Current app state: {app_client1.get_application_state()}")
    # Fund the app account with 1 algo
    app_client1.fund(1 * consts.algo)

    # Try calling set nickname from both accounts after opting in
    app_client1.opt_in()
    app_client1.call(ClientExample.set_nick, nick="first")

    print(f"Current app state: {app_client1.get_application_state()}")
    # Create copies of the app client with specific signer, _after_ we've created and set the app id
    app_client2 = app_client1.prepare(signer=acct2.signer)
    print(f"Current app state: {app_client1.get_application_state()}")

    # Try calling without opting in
    try:
        app_client2.call(ClientExample.set_nick, nick="second")
    except LogicException as e:
        print(f"\n{e}\n")

    app_client2.opt_in()
    app_client2.call(ClientExample.set_nick, nick="second")

    # Get the local state for each account
    print(app_client1.get_account_state())
    print(app_client2.get_account_state())

    # Get the global state
    print(f"Current app state: {app_client1.get_application_state()}")

    try:
        app_client2.call(ClientExample.set_manager, new_manager=acct2.address)
        print("Shouldn't get here")
    except LogicException as e:
        print("Failed as expected, only addr1 should be authorized to set the manager")
        print(f"\n{e}\n")

    # Have addr1 set the manager to addr2
    app_client1.call(ClientExample.set_manager, new_manager=acct2.address)
    print(f"Current app state: {app_client1.get_application_state()}")

    print(app_client1.get_application_state())
    # and back
    app_client2.call(ClientExample.set_manager, new_manager=acct1.address)
    print(f"Current app state: {app_client1.get_application_state()}")

    ## Create a new client that just sets the app id we wish to interact with
    app_client3 = ApplicationClient(
        client=sandbox.get_algod_client(),
        app=ClientExample(),
        signer=acct1.signer,
        app_id=app_client1.app_id,
    )

    try:
        app_client1.close_out()
        app_client2.close_out()
        app_client3.delete()
    except Exception as e:
        print(e)


if __name__ == "__main__":
    demo()



import copy

import itertools
import pytest
import typing

from algosdk.atomic_transaction_composer import (
    AtomicTransactionComposer,
    TransactionWithSigner,
    AccountTransactionSigner,
    abi,
)
from algosdk.future import transaction
from algosdk.v2client.algod import AlgodClient
from algosdk.encoding import decode_address
from beaker import client, sandbox, testing, consts, decorators
from beaker.client.application_client import ApplicationClient, ProgramAssertion
from beaker.client.logic_error import LogicException

from .amm import ConstantProductAMM, ConstantProductAMMErrors

accts = sandbox.get_accounts()
algod_client: AlgodClient = sandbox.get_algod_client()

TOTAL_POOL_TOKENS = 10000000000
TOTAL_ASSET_TOKENS = 10000000000


AcctInfo = tuple[str, str, AccountTransactionSigner]
AssertTestCase = tuple[
    str,
    abi.Method | decorators.HandlerFunc,
    dict[str, typing.Any],
    client.ApplicationClient,
]


@pytest.fixture(scope="session")
def creator_acct() -> AcctInfo:
    return accts[0].address, accts[0].private_key, accts[0].signer


@pytest.fixture(scope="session")
def user_acct() -> AcctInfo:
    return accts[1].address, accts[1].private_key, accts[1].signer


@pytest.fixture(scope="session")
def assets(creator_acct: AcctInfo, user_acct: AcctInfo) -> tuple[int, int]:
    addr, sk, _ = creator_acct
    user_addr, _, user_signer = user_acct

    sp = algod_client.suggested_params()
    txns: list[transaction.Transaction] = transaction.assign_group_id(
        [
            transaction.AssetCreateTxn(
                addr,
                sp,
                TOTAL_ASSET_TOKENS,
                0,
                False,
                asset_name="asset a",
                unit_name="A",
            ),
            transaction.AssetCreateTxn(
                addr,
                sp,
                TOTAL_ASSET_TOKENS,
                0,
                False,
                asset_name="asset b",
                unit_name="B",
            ),
        ]
    )
    algod_client.send_transactions([txn.sign(sk) for txn in txns])
    results = [
        transaction.wait_for_confirmation(algod_client, txid, 4)
        for txid in [t.get_txid() for t in txns]
    ]
    a_asset, b_asset = results[0]["asset-index"], results[1]["asset-index"]

    # Send some to the user account just to have them
    _opt_in_to_token(user_addr, user_signer, a_asset)
    _opt_in_to_token(user_addr, user_signer, b_asset)
    send_to_user_txns: list[transaction.Transaction] = transaction.assign_group_id(
        [
            transaction.AssetTransferTxn(
                addr, sp, user_addr, TOTAL_ASSET_TOKENS // 10, a_asset
            ),
            transaction.AssetTransferTxn(
                addr, sp, user_addr, TOTAL_ASSET_TOKENS // 10, b_asset
            ),
        ]
    )
    algod_client.send_transactions([txn.sign(sk) for txn in send_to_user_txns])

    return (a_asset, b_asset)


@pytest.fixture(scope="session")
def creator_app_client(creator_acct: AcctInfo) -> client.ApplicationClient:
    _, _, signer = creator_acct
    app = ConstantProductAMM()
    app_client = client.ApplicationClient(algod_client, app, signer=signer)
    return app_client


def test_app_create(creator_app_client: client.ApplicationClient):
    creator_app_client.create()
    app_state = creator_app_client.get_application_state()
    sender = creator_app_client.get_sender()

    assert app_state[ConstantProductAMM.governor.str_key()] == _addr_to_hex(
        sender
    ), "The governor should be my address"
    assert app_state[ConstantProductAMM.ratio.str_key()] == 0, "The ratio should be 0"


def minimum_fee_for_txn_count(
    sp: transaction.SuggestedParams, txn_count: int
) -> transaction.SuggestedParams:
    """
    Configures transaction fee _without_ considering network congestion.

    Since the function does not account for network congestion, do _not_ use the function as-is in a production use-case.
    """
    s = copy.deepcopy(sp)
    s.flat_fee = True
    s.fee = transaction.constants.min_txn_fee * txn_count
    return s


def assert_app_algo_balance(c: client.ApplicationClient, expected_algos: int):
    """
    Verifies the app's algo balance is not unexpectedly drained during app interaction (e.g. paying inner transaction fees).

    Due to the presence of rewards, the assertion tolerates actual > expected for small positive differences.
    """
    xs = testing.get_balances(c.client, [c.app_addr])
    assert c.app_addr in xs
    assert 0 in xs[c.app_addr]
    actual_algos = xs[c.app_addr][0]

    # Before accounting for rewards, confirm algos were not drained.
    assert actual_algos >= expected_algos

    # Account for rewards. 0 in devmode
    micro_algos_tolerance = 10
    assert actual_algos - expected_algos <= micro_algos_tolerance


app_algo_balance: typing.Final = consts.algo * 10


def build_set_governor_transaction(new_governor: str) -> dict[str, typing.Any]:
    return {"new_governor": new_governor}


def build_boostrap_transaction(
    app_client: client.ApplicationClient, assets: tuple[int, int]
) -> dict[str, typing.Any]:

    app_addr, addr, signer = (
        app_client.app_addr,
        app_client.sender,
        app_client.signer,
    )

    asset_a, asset_b = assets
    sp = app_client.client.suggested_params()

    return {
        "seed": TransactionWithSigner(
            txn=transaction.PaymentTxn(
                addr,
                sp,
                app_addr,
                app_algo_balance,
            ),
            signer=signer,
        ),
        "a_asset": asset_a,
        "b_asset": asset_b,
        "suggested_params": minimum_fee_for_txn_count(sp, 4),
    }


def build_mint_transaction(
    app_client: client.ApplicationClient,
    assets: tuple[int, int],
    pool_asset: int,
    a_amount: int,
    b_amount: int,
):

    app_addr, addr, signer = (
        app_client.app_addr,
        app_client.sender,
        app_client.signer,
    )

    a_asset, b_asset = assets
    sp = app_client.get_suggested_params()

    return {
        "suggested_params": minimum_fee_for_txn_count(sp, 2),
        "a_xfer": TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, a_amount, a_asset),
            signer=signer,
        ),
        "b_xfer": TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, b_amount, b_asset),
            signer=signer,
        ),
        "pool_asset": pool_asset,
        "a_asset": a_asset,
        "b_asset": b_asset,
    }


def build_burn_transaction(
    app_client: ApplicationClient,
    assets: tuple[int, int],
    pool_asset: int,
    burn_amt: int,
):

    app_addr, addr, signer = (
        app_client.app_addr,
        app_client.sender,
        app_client.signer,
    )

    sp = app_client.get_suggested_params()
    a_asset, b_asset = assets

    return {
        "suggested_params": minimum_fee_for_txn_count(sp, 3),
        "pool_xfer": TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, burn_amt, pool_asset),
            signer=signer,
        ),
        "pool_asset": pool_asset,
        "a_asset": a_asset,
        "b_asset": b_asset,
    }


def build_swap_transaction(
    app_client: ApplicationClient,
    assets: tuple[int, int],
    swap_amt: int,
    swap_asset: int = None,
):
    app_addr, addr, signer = (
        app_client.app_addr,
        app_client.sender,
        app_client.signer,
    )

    sp = app_client.get_suggested_params()
    a_asset, b_asset = assets

    if swap_asset is None:
        swap_asset = a_asset

    return {
        "suggested_params": minimum_fee_for_txn_count(sp, 2),
        "swap_xfer": TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, app_addr, swap_amt, swap_asset),
            signer=signer,
        ),
        "a_asset": a_asset,
        "b_asset": b_asset,
    }


def test_app_set_governor(
    creator_app_client: client.ApplicationClient, user_acct: AcctInfo
):
    creator_addr, _ = creator_app_client.sender, creator_app_client.app_addr

    user_addr, _, user_signer = user_acct

    state_before = creator_app_client.get_application_state()

    assert creator_addr is not None
    assert state_before[ConstantProductAMM.governor.str_key()] == _addr_to_hex(
        creator_addr
    )

    # Set the new gov
    creator_app_client.call(
        ConstantProductAMM.set_governor,
        **build_set_governor_transaction(user_addr),
    )

    state_after = creator_app_client.get_application_state()
    assert state_after[ConstantProductAMM.governor.str_key()] == _addr_to_hex(user_addr)

    user_client = creator_app_client.prepare(signer=user_signer)
    # Return state to old gov
    user_client.call(
        ConstantProductAMM.set_governor,
        **build_set_governor_transaction(creator_addr),
    )

    state_after_revert = creator_app_client.get_application_state()
    assert state_after_revert[ConstantProductAMM.governor.str_key()] == _addr_to_hex(
        creator_addr
    )


def test_app_bootstrap(
    creator_app_client: client.ApplicationClient, assets: tuple[int, int]
):

    app_addr = creator_app_client.app_addr
    asset_a, asset_b = assets

    # Bootstrap to create pool token and set global state
    result = creator_app_client.call(
        ConstantProductAMM.bootstrap,
        **build_boostrap_transaction(creator_app_client, assets),
    )

    assert_app_algo_balance(creator_app_client, app_algo_balance)

    pool_token = result.return_value
    assert pool_token > 0, "We should have created a pool token with asset id>0"

    # Check pool token params
    token_info = creator_app_client.client.asset_info(pool_token)
    assert token_info["params"]["name"] == "DPT-A-B"
    assert token_info["params"]["total"] == TOTAL_POOL_TOKENS
    assert token_info["params"]["reserve"] == app_addr
    assert token_info["params"]["manager"] == app_addr
    assert token_info["params"]["creator"] == app_addr

    # Make sure we're opted in
    ai = creator_app_client.get_application_account_info()
    assert len(ai["assets"]) == 3, "Should have 3 assets, A/B/Pool"

    # Make sure our state is updated
    app_state = creator_app_client.get_application_state()
    assert app_state[ConstantProductAMM.pool_token.str_key()] == pool_token
    assert app_state[ConstantProductAMM.asset_a.str_key()] == asset_a
    assert app_state[ConstantProductAMM.asset_b.str_key()] == asset_b


def test_app_fund(creator_app_client: ApplicationClient):
    app_addr, addr, signer = (
        creator_app_client.app_addr,
        creator_app_client.sender,
        creator_app_client.signer,
    )

    pool_asset, a_asset, b_asset = _get_tokens_from_state(creator_app_client)

    assert addr
    _opt_in_to_token(addr, signer, pool_asset)

    balance_accts = [app_addr, addr]
    balances_before = testing.get_balances(creator_app_client.client, balance_accts)

    a_amount = 10000
    b_amount = 3000

    creator_app_client.call(
        ConstantProductAMM.mint,
        **build_mint_transaction(
            creator_app_client, (a_asset, b_asset), pool_asset, a_amount, b_amount
        ),
    )

    balances_after = testing.get_balances(creator_app_client.client, balance_accts)
    balance_deltas = testing.get_deltas(balances_before, balances_after)

    assert balance_deltas[app_addr][a_asset] == a_amount
    assert balance_deltas[app_addr][b_asset] == b_amount
    assert_app_algo_balance(creator_app_client, app_algo_balance)

    expected_pool_tokens = int((a_amount * b_amount) ** 0.5 - ConstantProductAMM._scale)
    assert balance_deltas[addr][pool_asset] == expected_pool_tokens

    ratio = _get_ratio_from_state(creator_app_client)
    expected_ratio = int((a_amount * ConstantProductAMM._scale) / b_amount)
    assert ratio == expected_ratio


def test_mint(creator_app_client: ApplicationClient):
    app_addr, addr = (
        creator_app_client.app_addr,
        creator_app_client.sender,
    )

    pool_asset, a_asset, b_asset = _get_tokens_from_state(creator_app_client)

    assert addr
    balances_before = testing.get_balances(creator_app_client.client, [app_addr, addr])

    ratio_before = _get_ratio_from_state(creator_app_client)

    a_amount = 40000
    b_amount = int(a_amount * ConstantProductAMM._scale / ratio_before)

    creator_app_client.call(
        ConstantProductAMM.mint,
        **build_mint_transaction(
            creator_app_client, (a_asset, b_asset), pool_asset, a_amount, b_amount
        ),
    )

    balances_after = testing.get_balances(creator_app_client.client, [app_addr, addr])
    balance_deltas = testing.get_deltas(balances_before, balances_after)

    # App got the right amount
    assert balance_deltas[app_addr][a_asset] == a_amount
    assert balance_deltas[app_addr][b_asset] == b_amount
    assert_app_algo_balance(creator_app_client, app_algo_balance)

    # We minted the correct amount of pool tokens
    issued = TOTAL_POOL_TOKENS - balances_before[app_addr][pool_asset]
    expected_pool_tokens = _get_tokens_to_mint(
        issued,
        a_amount,
        balances_before[app_addr][a_asset],
        b_amount,
        balances_before[app_addr][b_asset],
        ConstantProductAMM._scale,
    )
    assert balance_deltas[addr][pool_asset] == int(expected_pool_tokens)

    # We updated the ratio accordingly
    actual_ratio = _get_ratio_from_state(creator_app_client)
    expected_ratio = _expect_ratio(
        balances_after[app_addr][a_asset], balances_after[app_addr][b_asset]
    )
    assert actual_ratio == expected_ratio


def test_burn(creator_app_client: ApplicationClient):
    app_addr, addr = (
        creator_app_client.app_addr,
        creator_app_client.sender,
    )

    pool_asset, a_asset, b_asset = _get_tokens_from_state(creator_app_client)

    assert addr
    balances_before = testing.get_balances(creator_app_client.client, [app_addr, addr])

    burn_amt = balances_before[addr][pool_asset] // 10

    creator_app_client.call(
        ConstantProductAMM.burn,
        **build_burn_transaction(
            creator_app_client, (a_asset, b_asset), pool_asset, burn_amt
        ),
    )

    balances_after = testing.get_balances(creator_app_client.client, [app_addr, addr])
    balances_delta = testing.get_deltas(balances_before, balances_after)

    assert balances_delta[app_addr][pool_asset] == burn_amt

    # We minted the correct amount of pool tokens
    issued = TOTAL_POOL_TOKENS - balances_before[app_addr][pool_asset]
    a_supply = balances_before[app_addr][a_asset]
    b_supply = balances_before[app_addr][b_asset]

    expected_a_tokens = _get_tokens_to_burn(a_supply, burn_amt, issued)
    assert balances_delta[addr][a_asset] == int(expected_a_tokens)

    expected_b_tokens = _get_tokens_to_burn(b_supply, burn_amt, issued)
    assert balances_delta[addr][b_asset] == int(expected_b_tokens)

    assert_app_algo_balance(creator_app_client, app_algo_balance)

    ratio_after = _get_ratio_from_state(creator_app_client)

    # Ratio should be identical?
    # assert ratio_before == ratio_after

    expected_ratio = _expect_ratio(
        balances_after[app_addr][a_asset], balances_after[app_addr][b_asset]
    )
    assert ratio_after == expected_ratio


def test_swap(creator_app_client: ApplicationClient):
    app_addr, addr = (
        creator_app_client.app_addr,
        creator_app_client.sender,
    )

    _, a_asset, b_asset = _get_tokens_from_state(creator_app_client)

    assert addr
    balances_before = testing.get_balances(creator_app_client.client, [app_addr, addr])

    swap_amt = balances_before[addr][a_asset] // 10
    creator_app_client.call(
        ConstantProductAMM.swap,
        **build_swap_transaction(creator_app_client, (a_asset, b_asset), swap_amt),
    )

    balances_after = testing.get_balances(creator_app_client.client, [app_addr, addr])
    balances_delta = testing.get_deltas(balances_before, balances_after)

    assert balances_delta[app_addr][a_asset] == swap_amt

    # We minted the correct amount of pool tokens
    a_supply = balances_before[app_addr][a_asset]
    b_supply = balances_before[app_addr][b_asset]

    expected_b_tokens = _get_tokens_to_swap(
        swap_amt, a_supply, b_supply, ConstantProductAMM._scale, ConstantProductAMM._fee
    )
    assert balances_delta[addr][b_asset] == int(expected_b_tokens)

    assert_app_algo_balance(creator_app_client, app_algo_balance)

    ratio_after = _get_ratio_from_state(creator_app_client)
    expected_ratio = _expect_ratio(
        balances_after[app_addr][a_asset], balances_after[app_addr][b_asset]
    )
    assert ratio_after == expected_ratio


all_assert_groups = ["governor", "bootstrap", "mint", "burn", "swap"]


@pytest.fixture(
    scope="session",
    params=all_assert_groups,
)
def grouped_assert_cases(
    request, creator_app_client: client.ApplicationClient, user_acct: AcctInfo
) -> list[AssertTestCase]:
    return _assert_cases(request.param, creator_app_client, user_acct)


@pytest.fixture(scope="session")
def all_assert_cases(
    creator_app_client: client.ApplicationClient, user_acct: AcctInfo
) -> list[AssertTestCase]:
    return _assert_cases("all", creator_app_client, user_acct)


def _assert_cases(
    group_key: str,
    creator_app_client: client.ApplicationClient,
    user_acct: AcctInfo,
) -> list[AssertTestCase]:
    def cases(
        m: abi.Method | decorators.HandlerFunc,
        xs: list[tuple[str, dict[str, typing.Any]]],
        client: client.ApplicationClient = creator_app_client,
    ) -> list[AssertTestCase]:
        return [(a, m, txn, client) for a, txn in xs]

    fake_addr, fake_pk, fake_signer = user_acct
    fake_client = creator_app_client.prepare(signer=fake_signer, sender=fake_addr)

    app_addr, addr, signer = (
        creator_app_client.app_addr,
        creator_app_client.sender,
        creator_app_client.signer,
    )

    pool_asset, a_asset, b_asset = _get_tokens_from_state(creator_app_client)
    assets = (a_asset, b_asset)

    _opt_in_to_token(fake_addr, fake_signer, pool_asset)

    sp = creator_app_client.client.suggested_params()

    def add_txn(
        d: dict[str, AtomicTransactionComposer], key: str
    ) -> dict[str, AtomicTransactionComposer]:
        d[key] = AtomicTransactionComposer().add_transaction(
            TransactionWithSigner(
                txn=transaction.PaymentTxn(addr, sp, addr, 0),
                signer=signer,
            )
        )
        return d

    def wrong_receiver(
        d: dict[str, TransactionWithSigner], key: str
    ) -> dict[str, TransactionWithSigner]:
        d[key].txn.receiver = addr
        return d

    def override_pay_amount(
        d: dict[str, TransactionWithSigner], key: str, amt: int
    ) -> dict[str, TransactionWithSigner]:
        d[key].txn.amt = amt
        return d

    def override_axfer_amount(
        d: dict[str, TransactionWithSigner], key: str, amt: int
    ) -> dict[str, TransactionWithSigner]:
        d[key].txn.amount = amt
        return d

    def override_axfer_asset(
        d: dict[str, TransactionWithSigner], key: str, override: int
    ) -> dict[str, TransactionWithSigner]:
        d[key].txn.index = override
        return d

    def set_governor_cases():
        def set_governor(new_gov: str):
            return build_set_governor_transaction(new_governor=new_gov)

        return cases(
            ConstantProductAMM.set_governor,
            [("unauthorized", set_governor(addr))],
            fake_client,
        )

    def bootstrap_cases():
        def bootstrap(
            app_client: client.ApplicationClient = creator_app_client,
            assets: tuple[int, int] = assets,
        ):
            return build_boostrap_transaction(app_client, assets)

        return cases(
            ConstantProductAMM.bootstrap,
            [
                (ConstantProductAMMErrors.GroupSizeNot2, add_txn(bootstrap(), "atc")),
                (
                    ConstantProductAMMErrors.ReceiverNotAppAddr,
                    wrong_receiver(bootstrap(), "seed"),
                ),
                (
                    ConstantProductAMMErrors.AmountLessThanMinimum,
                    override_pay_amount(bootstrap(), "seed", int(consts.algo * 0.29)),
                ),
                (
                    ConstantProductAMMErrors.AssetIdsIncorrect,
                    bootstrap(assets=(b_asset, b_asset)),
                ),
            ],
        ) + cases(
            ConstantProductAMM.bootstrap, [("unauthorized", bootstrap())], fake_client
        )

    def mint_cases():
        a_amt = 100000
        b_amt = a_amt // 10

        def mint(
            app_client: client.ApplicationClient = creator_app_client,
            assets: tuple[int, int] = assets,
            pool_asset: int = pool_asset,
            a_amount: int = a_amt,
            b_amount: int = b_amt,
        ):
            return build_mint_transaction(
                app_client, assets, pool_asset, a_amount, b_amount
            )

        well_formed_mint = cases(
            ConstantProductAMM.mint,
            [
                (
                    ConstantProductAMMErrors.AssetAIncorrect,
                    mint(assets=(b_asset, b_asset)),
                ),
                (
                    ConstantProductAMMErrors.AssetBIncorrect,
                    mint(assets=(a_asset, a_asset)),
                ),
                (
                    ConstantProductAMMErrors.AssetPoolIncorrect,
                    mint(pool_asset=a_asset),
                ),
            ],
        ) + cases(
            ConstantProductAMM.mint,
            [(ConstantProductAMMErrors.SenderInvalid, mint())],
            fake_client,
        )

        def valid_asset_xfer(key: str):
            if key not in ["a_xfer", "b_xfer"]:
                raise Exception(f"Unexpected {key=}")

            return [
                (
                    ConstantProductAMMErrors.ReceiverNotAppAddr,
                    wrong_receiver(mint(), key),
                ),
                (
                    ConstantProductAMMErrors.AssetAIncorrect
                    if key == "a_xfer"
                    else ConstantProductAMMErrors.AssetBIncorrect,
                    override_axfer_asset(
                        mint(), key, b_asset if key == "a_xfer" else a_asset
                    ),
                ),
                (
                    ConstantProductAMMErrors.AmountLessThanMinimum,
                    override_axfer_amount(mint(), key, 0),
                ),
                (
                    ConstantProductAMMErrors.SendAmountTooLow,
                    override_axfer_amount(mint(), key, 1),
                ),
            ]

        valid_asset_a_xfer = cases(ConstantProductAMM.mint, valid_asset_xfer("a_xfer"))

        valid_asset_b_xfer = cases(ConstantProductAMM.mint, valid_asset_xfer("b_xfer"))

        return well_formed_mint + valid_asset_a_xfer + valid_asset_b_xfer

    def burn_cases():
        def burn(
            app_client: client.ApplicationClient = creator_app_client,
            assets: tuple[int, int] = assets,
            pool_asset: int = pool_asset,
            burn_amt: int = 1,
        ):
            return build_burn_transaction(app_client, assets, pool_asset, burn_amt)

        well_formed_burn = cases(
            ConstantProductAMM.burn,
            [
                (ConstantProductAMMErrors.AssetPoolIncorrect, burn(pool_asset=a_asset)),
                (
                    ConstantProductAMMErrors.AssetAIncorrect,
                    burn(assets=(b_asset, b_asset)),
                ),
                (
                    ConstantProductAMMErrors.AssetBIncorrect,
                    burn(assets=(a_asset, a_asset)),
                ),
            ],
        )

        valid_pool_xfer = cases(
            ConstantProductAMM.burn,
            [
                (
                    ConstantProductAMMErrors.ReceiverNotAppAddr,
                    wrong_receiver(burn(), "pool_xfer"),
                ),
                (ConstantProductAMMErrors.AmountLessThanMinimum, burn(burn_amt=0)),
                (
                    ConstantProductAMMErrors.AssetPoolIncorrect,
                    override_axfer_asset(burn(), "pool_xfer", a_asset),
                ),
            ],
        ) + cases(
            ConstantProductAMM.burn,
            [(ConstantProductAMMErrors.SenderInvalid, burn())],
            fake_client,
        )

        return well_formed_burn + valid_pool_xfer

    def swap_cases():
        def swap(
            app_client: client.ApplicationClient = creator_app_client,
            assets: tuple[int, int] = assets,
            swap_amt: int = 1,
            swap_asset: int = a_asset,
        ):
            return build_swap_transaction(app_client, assets, swap_amt, swap_asset)

        well_formed_swap = cases(
            ConstantProductAMM.swap,
            [
                (
                    ConstantProductAMMErrors.AssetAIncorrect,
                    swap(assets=(b_asset, b_asset)),
                ),
                (
                    ConstantProductAMMErrors.AssetBIncorrect,
                    swap(assets=(a_asset, a_asset)),
                ),
            ],
        )

        valid_swap_xfer = cases(
            ConstantProductAMM.swap,
            [
                (ConstantProductAMMErrors.AmountLessThanMinimum, swap(swap_amt=0)),
                (
                    ConstantProductAMMErrors.SendAmountTooLow,
                    swap(swap_amt=1),
                ),
                (
                    ConstantProductAMMErrors.AssetIdsIncorrect,
                    swap(swap_amt=0, swap_asset=pool_asset),
                ),
            ],
        ) + cases(
            ConstantProductAMM.swap,
            [(ConstantProductAMMErrors.SenderInvalid, swap())],
            fake_client,
        )

        return well_formed_swap + valid_swap_xfer

    key_to_group: dict[str, list[AssertTestCase]] = {
        "governor": set_governor_cases(),
        "bootstrap": bootstrap_cases(),
        "mint": mint_cases(),
        "burn": burn_cases(),
        "swap": swap_cases(),
    }

    # Sanity check - Confirm additions to `key_to_group` are added to fixture parameterization.
    assert sorted(key_to_group.keys()) == sorted(all_assert_groups)

    all_cases = list(itertools.chain.from_iterable(key_to_group.values()))
    return all_cases if group_key == "all" else key_to_group[group_key]


def test_approval_asserts(grouped_assert_cases: list[AssertTestCase]):
    """
    Confirms each logical grouping of assertions raises the expected error message.
    """
    for msg, method, kwargs, app_client in grouped_assert_cases:
        with pytest.raises(LogicException, match=msg):
            app_client.call(method, **kwargs)


def test_approval_assert_coverage(
    all_assert_cases: list[AssertTestCase], creator_app_client: client.ApplicationClient
):
    """
    Confirms `test_approval_asserts` exercises all app approval asserts.

    If `test_approval_asserts` passes and this test fails, it implies some asserts are _not_ tested.
    """

    all_asserts: dict[int, ProgramAssertion] = creator_app_client.approval_asserts  # type: ignore[assignment]

    for msg, method, kwargs, app_client in all_assert_cases:
        with pytest.raises(LogicException, match=msg):
            try:
                app_client.call(method, **kwargs)
            except LogicException as e:
                if e.pc in all_asserts:
                    del all_asserts[e.pc]
                raise e

    assert len(all_asserts) == 0


def _get_tokens_to_mint(
    issued: int, a_amt: int, a_supply: int, b_amt: int, b_supply: int, scale: int
) -> int:
    a_ratio = (a_amt * scale) / a_supply
    b_ratio = (b_amt * scale) / b_supply

    if a_ratio < b_ratio:
        return int((a_ratio * issued) / scale)

    return int((b_ratio * issued) / scale)


def _get_tokens_to_swap(in_amount, in_supply, out_supply, scale, fee) -> int:
    factor = scale - fee
    return int(
        (in_amount * factor * out_supply) / ((in_supply * scale) + (in_amount * factor))
    )


def _get_tokens_to_burn(asset_supply, burn_amount, pool_issued):
    return int((asset_supply * burn_amount) / pool_issued)


def _get_ratio_from_state(creator_app_client: ApplicationClient):
    app_state = creator_app_client.get_application_state()
    return app_state[ConstantProductAMM.ratio.str_key()]


def _get_tokens_from_state(
    creator_app_client: ApplicationClient,
) -> tuple[int, int, int]:
    app_state = creator_app_client.get_application_state()
    return (
        int(app_state[ConstantProductAMM.pool_token.str_key()]),
        int(app_state[ConstantProductAMM.asset_a.str_key()]),
        int(app_state[ConstantProductAMM.asset_b.str_key()]),
    )


def _expect_ratio(a_sup, b_sup):
    return int((a_sup * ConstantProductAMM._scale) / b_sup)


def _opt_in_to_token(addr: str, signer: AccountTransactionSigner, id: int):
    sp = algod_client.suggested_params()
    atc = AtomicTransactionComposer()
    atc.add_transaction(
        TransactionWithSigner(
            txn=transaction.AssetTransferTxn(addr, sp, addr, 0, id),
            signer=signer,
        )
    )
    atc.execute(algod_client, 2)


def _addr_to_hex(addr: str) -> str:
    return decode_address(addr).hex()



from typing import Final
from pyteal import (
    abi,
    pragma,
    TealType,
    Bytes,
    Global,
    Expr,
    Int,
    Seq,
    Assert,
    Txn,
    And,
    ScratchVar,
    AssetHolding,
    AssetParam,
    WideRatio,
    If,
    Or,
    InnerTxn,
    InnerTxnBuilder,
    TxnField,
    Concat,
    TxnType,
    Sqrt,
)

from beaker import (
    consts,
    ApplicationStateValue,
    Application,
    Authorize,
    external,
    create,
    internal,
)

# WARNING: This code is provided for example only. Do NOT deploy to mainnet.

pragma(compiler_version="^0.20.1")


def commented_assert(conditions: list[tuple[Expr, str]]) -> list[Expr]:
    return [Assert(cond, comment=cmt) for cond, cmt in conditions]


class ConstantProductAMMErrors:
    GroupSizeNot2 = "group size not 2"
    ReceiverNotAppAddr = "receiver not app address"
    AmountLessThanMinimum = "amount minimum not met"
    AssetIdsIncorrect = "asset a or asset b incorrect"
    AssetAIncorrect = "asset a incorrect"
    AssetBIncorrect = "asset b incorrect"
    AssetPoolIncorrect = "asset pool incorrect"
    SenderInvalid = "invalid sender"
    MissingBalances = "missing required balances"
    SendAmountTooLow = "outgoing amount too low"


class ConstantProductAMM(Application):

    # Declare Application state, marking `Final` here so the python class var doesn't get changed
    # Marking a var `Final` does _not_ change anything at the AVM level
    governor: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.bytes,
        key=Bytes("g"),
        default=Global.creator_address(),
        descr="The current governor of this contract, allowed to do admin type actions",
    )
    asset_a: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        key=Bytes("a"),
        static=True,
        descr="The asset id of asset A",
    )
    asset_b: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        key=Bytes("b"),
        static=True,
        descr="The asset id of asset B",
    )
    pool_token: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        key=Bytes("p"),
        static=True,
        descr="The asset id of the Pool Token, used to track share of pool the holder may recover",
    )
    ratio: Final[ApplicationStateValue] = ApplicationStateValue(
        stack_type=TealType.uint64,
        key=Bytes("r"),
        descr="The ratio between assets (A/B)*Scale",
    )

    ##############
    # Constants
    ##############

    # Total supply of the pool tokens
    _total_supply: Final[int] = int(1e10)
    total_supply: Final[Expr] = Int(_total_supply)
    # scale helps with precision when doing computation for
    # the number of tokens to transfer
    _scale: Final[int] = 1000
    scale: Final[Expr] = Int(_scale)
    # Fee for swaps, 5 represents 0.5% ((fee / scale)*100)
    _fee: Final[int] = 5
    fee: Final[Expr] = Int(_fee)

    ##############
    # Administrative Actions
    ##############

    # Call this only on create
    @create
    def create(self):
        return self.initialize_application_state()

    # Only the account set in app_state.governor may call this method
    @external(authorize=Authorize.only(governor))
    def set_governor(self, new_governor: abi.Account):
        """sets the governor of the contract, may only be called by the current governor"""
        return self.governor.set(new_governor.address())

    # Only the account set in app_state.governor may call this method
    @external(authorize=Authorize.only(governor))
    def bootstrap(
        self,
        seed: abi.PaymentTransaction,
        a_asset: abi.Asset,
        b_asset: abi.Asset,
        *,
        output: abi.Uint64,
    ):
        """bootstraps the contract by opting into the assets and creating the pool token.

        Note this method will fail if it is attempted more than once on the same contract since the assets and pool token
        application state values are marked as static and cannot be overridden.

        Args:
            seed: Initial Payment transaction to the app account so it can opt in to assets and create pool token.
            a_asset: One of the two assets this pool should allow swapping between.
            b_asset: The other of the two assets this pool should allow swapping between.

        Returns:
            The asset id of the pool token created.
        """

        well_formed_bootstrap = [
            (Global.group_size() == Int(2), ConstantProductAMMErrors.GroupSizeNot2),
            (
                seed.get().receiver() == self.address,
                ConstantProductAMMErrors.ReceiverNotAppAddr,
            ),
            (
                seed.get().amount() >= consts.Algos(0.3),
                ConstantProductAMMErrors.AmountLessThanMinimum,
            ),
            (
                a_asset.asset_id() < b_asset.asset_id(),
                ConstantProductAMMErrors.AssetIdsIncorrect,
            ),
        ]

        return Seq(
            *commented_assert(well_formed_bootstrap),
            self.asset_a.set(a_asset.asset_id()),
            self.asset_b.set(b_asset.asset_id()),
            self.pool_token.set(
                self.do_create_pool_token(
                    self.asset_a,
                    self.asset_b,
                ),
            ),
            self.do_opt_in(self.asset_a),
            self.do_opt_in(self.asset_b),
            output.set(self.pool_token),
        )

    ##############
    # AMM specific methods for mint/burn/swap
    ##############

    @external
    def mint(
        self,
        a_xfer: abi.AssetTransferTransaction,
        b_xfer: abi.AssetTransferTransaction,
        pool_asset: abi.Asset = pool_token,  # type: ignore[assignment]
        a_asset: abi.Asset = asset_a,  # type: ignore[assignment]
        b_asset: abi.Asset = asset_b,  # type: ignore[assignment]
    ):
        """mint pool tokens given some amount of asset A and asset B.

        Given some amount of Asset A and Asset B in the transfers, mint some number of pool tokens commensurate with
        the pools current balance and circulating supply of pool tokens.

        Args:
            a_xfer: Asset Transfer Transaction of asset A as a deposit to the pool in exchange for pool tokens.
            b_xfer: Asset Transfer Transaction of asset B as a deposit to the pool in exchange for pool tokens.
            pool_asset: The asset ID of the pool token so that we may distribute it.
            a_asset: The asset ID of the Asset A so that we may inspect our balance.
            b_asset: The asset ID of the Asset B so that we may inspect our balance.
        """

        well_formed_mint = [
            (
                a_asset.asset_id() == self.asset_a,
                ConstantProductAMMErrors.AssetAIncorrect,
            ),
            (
                b_asset.asset_id() == self.asset_b,
                ConstantProductAMMErrors.AssetBIncorrect,
            ),
            (
                pool_asset.asset_id() == self.pool_token,
                ConstantProductAMMErrors.AssetPoolIncorrect,
            ),
            (
                And(
                    a_xfer.get().sender() == Txn.sender(),
                    b_xfer.get().sender() == Txn.sender(),
                ),
                ConstantProductAMMErrors.SenderInvalid,
            ),
        ]

        valid_asset_a_xfer = [
            (
                a_xfer.get().asset_receiver() == self.address,
                ConstantProductAMMErrors.ReceiverNotAppAddr,
            ),
            (
                a_xfer.get().xfer_asset() == self.asset_a,
                ConstantProductAMMErrors.AssetAIncorrect,
            ),
            (
                a_xfer.get().asset_amount() > Int(0),
                ConstantProductAMMErrors.AmountLessThanMinimum,
            ),
        ]

        valid_asset_b_xfer = [
            (
                b_xfer.get().asset_receiver() == self.address,
                ConstantProductAMMErrors.ReceiverNotAppAddr,
            ),
            (
                b_xfer.get().xfer_asset() == self.asset_b,
                ConstantProductAMMErrors.AssetBIncorrect,
            ),
            (
                b_xfer.get().asset_amount() > Int(0),
                ConstantProductAMMErrors.AmountLessThanMinimum,
            ),
        ]

        return Seq(
            # Check that the transaction is constructed correctly
            *commented_assert(
                well_formed_mint + valid_asset_a_xfer + valid_asset_b_xfer
            ),
            # Check that we have these things
            pool_bal := pool_asset.holding(self.address).balance(),
            a_bal := a_asset.holding(self.address).balance(),
            b_bal := b_asset.holding(self.address).balance(),
            Assert(
                pool_bal.hasValue(),
                a_bal.hasValue(),
                b_bal.hasValue(),
            ),
            (to_mint := ScratchVar()).store(
                If(
                    And(
                        a_bal.value() == a_xfer.get().asset_amount(),
                        b_bal.value() == b_xfer.get().asset_amount(),
                    ),
                    # This is the first time we've been called
                    # we use a different formula to mint tokens
                    self.tokens_to_mint_initial(
                        a_xfer.get().asset_amount(), b_xfer.get().asset_amount()
                    ),
                    # Normal mint
                    self.tokens_to_mint(
                        self.total_supply - pool_bal.value(),
                        a_bal.value() - a_xfer.get().asset_amount(),
                        b_bal.value() - b_xfer.get().asset_amount(),
                        a_xfer.get().asset_amount(),
                        b_xfer.get().asset_amount(),
                    ),
                )
            ),
            Assert(
                to_mint.load() > Int(0),
                comment=ConstantProductAMMErrors.SendAmountTooLow,
            ),
            # mint tokens
            self.do_axfer(Txn.sender(), self.pool_token, to_mint.load()),
            self.ratio.set(self.compute_ratio()),
        )

    @external
    def burn(
        self,
        pool_xfer: abi.AssetTransferTransaction,
        pool_asset: abi.Asset = pool_token,  # type: ignore[assignment]
        a_asset: abi.Asset = asset_a,  # type: ignore[assignment]
        b_asset: abi.Asset = asset_b,  # type: ignore[assignment]
    ):
        """burn pool tokens to get back some amount of asset A and asset B

        Args:
            pool_xfer: Asset Transfer Transaction of the pool token for the amount the sender wishes to redeem
            pool_asset: Asset ID of the pool token so we may inspect balance.
            a_asset: Asset ID of Asset A so we may inspect balance and distribute it
            b_asset: Asset ID of Asset B so we may inspect balance and distribute it
        """

        well_formed_burn = [
            (
                pool_asset.asset_id() == self.pool_token,
                ConstantProductAMMErrors.AssetPoolIncorrect,
            ),
            (
                a_asset.asset_id() == self.asset_a,
                ConstantProductAMMErrors.AssetAIncorrect,
            ),
            (
                b_asset.asset_id() == self.asset_b,
                ConstantProductAMMErrors.AssetBIncorrect,
            ),
        ]

        valid_pool_xfer = [
            (
                pool_xfer.get().asset_receiver() == self.address,
                ConstantProductAMMErrors.ReceiverNotAppAddr,
            ),
            (
                pool_xfer.get().asset_amount() > Int(0),
                ConstantProductAMMErrors.AmountLessThanMinimum,
            ),
            (
                pool_xfer.get().xfer_asset() == self.pool_token,
                ConstantProductAMMErrors.AssetPoolIncorrect,
            ),
            (
                pool_xfer.get().sender() == Txn.sender(),
                ConstantProductAMMErrors.SenderInvalid,
            ),
        ]

        return Seq(
            *commented_assert(well_formed_burn + valid_pool_xfer),
            pool_bal := pool_asset.holding(self.address).balance(),
            a_bal := a_asset.holding(self.address).balance(),
            b_bal := b_asset.holding(self.address).balance(),
            Assert(
                pool_bal.hasValue(),
                a_bal.hasValue(),
                b_bal.hasValue(),
            ),
            # Get the total number of tokens issued (prior to receiving the current axfer of pool tokens)
            (issued := ScratchVar()).store(
                self.total_supply - (pool_bal.value() - pool_xfer.get().asset_amount())
            ),
            (a_amt := ScratchVar()).store(
                self.tokens_to_burn(
                    issued.load(),
                    a_bal.value(),
                    pool_xfer.get().asset_amount(),
                )
            ),
            (b_amt := ScratchVar()).store(
                self.tokens_to_burn(
                    issued.load(),
                    b_bal.value(),
                    pool_xfer.get().asset_amount(),
                )
            ),
            # Send back commensurate amt of a
            self.do_axfer(
                Txn.sender(),
                self.asset_a,
                a_amt.load(),
            ),
            # Send back commensurate amt of b
            self.do_axfer(
                Txn.sender(),
                self.asset_b,
                b_amt.load(),
            ),
            self.ratio.set(self.compute_ratio()),
        )

    @external
    def swap(
        self,
        swap_xfer: abi.AssetTransferTransaction,
        a_asset: abi.Asset = asset_a,  # type: ignore[assignment]
        b_asset: abi.Asset = asset_b,  # type: ignore[assignment]
    ):
        """Swap some amount of either asset A or asset B for the other

        Args:
            swap_xfer: Asset Transfer Transaction of either Asset A or Asset B
            a_asset: Asset ID of asset A so we may inspect balance and possibly transfer it
            b_asset: Asset ID of asset B so we may inspect balance and possibly transfer it
        """
        well_formed_swap = [
            (
                a_asset.asset_id() == self.asset_a,
                ConstantProductAMMErrors.AssetAIncorrect,
            ),
            (
                b_asset.asset_id() == self.asset_b,
                ConstantProductAMMErrors.AssetBIncorrect,
            ),
        ]

        valid_swap_xfer = [
            (
                Or(
                    swap_xfer.get().xfer_asset() == self.asset_a,
                    swap_xfer.get().xfer_asset() == self.asset_b,
                ),
                ConstantProductAMMErrors.AssetIdsIncorrect,
            ),
            (
                swap_xfer.get().asset_amount() > Int(0),
                ConstantProductAMMErrors.AmountLessThanMinimum,
            ),
            (
                swap_xfer.get().sender() == Txn.sender(),
                ConstantProductAMMErrors.SenderInvalid,
            ),
        ]

        out_id = If(
            swap_xfer.get().xfer_asset() == self.asset_a,
            self.asset_b,
            self.asset_a,
        )
        in_id = swap_xfer.get().xfer_asset()

        return Seq(
            *commented_assert(well_formed_swap + valid_swap_xfer),
            in_sup := AssetHolding.balance(self.address, in_id),
            out_sup := AssetHolding.balance(self.address, out_id),
            Assert(
                in_sup.hasValue(),
                out_sup.hasValue(),
            ),
            (to_swap := ScratchVar()).store(
                self.tokens_to_swap(
                    swap_xfer.get().asset_amount(),
                    in_sup.value() - swap_xfer.get().asset_amount(),
                    out_sup.value(),
                )
            ),
            Assert(
                to_swap.load() > Int(0),
                comment=ConstantProductAMMErrors.SendAmountTooLow,
            ),
            self.do_axfer(
                Txn.sender(),
                out_id,
                to_swap.load(),
            ),
            self.ratio.set(self.compute_ratio()),
        )

    ##############
    # Mathy methods
    ##############

    # Notes:
    #   1) During arithmetic operations, depending on the inputs, these methods may overflow
    #   the max uint64 value. This will cause the program to immediately terminate.
    #
    #   Care should be taken to fully understand the limitations of these functions and if
    #   required should be swapped out for the appropriate byte math operations.
    #
    #   2) When doing division, any remainder is truncated from the result.
    #
    #   Care should be taken  to ensure that _when_ the truncation happens,
    #   it does so in favor of the contract. This is a subtle security issue that,
    #   if mishandled, could cause the balance of the contract to be drained.

    @internal(TealType.uint64)
    def tokens_to_mint(self, issued, a_supply, b_supply, a_amount, b_amount):
        return Seq(
            (a_rat := ScratchVar()).store(
                WideRatio([a_amount, self.scale], [a_supply])
            ),
            (b_rat := ScratchVar()).store(
                WideRatio([b_amount, self.scale], [b_supply])
            ),
            WideRatio(
                [If(a_rat.load() < b_rat.load(), a_rat.load(), b_rat.load()), issued],
                [self.scale],
            ),
        )

    @internal(TealType.uint64)
    def tokens_to_mint_initial(self, a_amount, b_amount):
        return Sqrt(a_amount * b_amount) - self.scale

    @internal(TealType.uint64)
    def tokens_to_burn(self, issued, supply, amount):
        return WideRatio([supply, amount], [issued])

    @internal(TealType.uint64)
    def tokens_to_swap(self, in_amount, in_supply, out_supply):
        factor = self.scale - self.fee
        return WideRatio(
            [in_amount, factor, out_supply],
            [(in_supply * self.scale) + (in_amount * factor)],
        )

    ##############
    # Utility methods for inner transactions
    ##############

    @internal(TealType.none)
    def do_axfer(self, rx, aid, amt):
        return InnerTxnBuilder.Execute(
            {
                TxnField.type_enum: TxnType.AssetTransfer,
                TxnField.xfer_asset: aid,
                TxnField.asset_amount: amt,
                TxnField.asset_receiver: rx,
                TxnField.fee: Int(0),
            }
        )

    @internal(TealType.none)
    def do_opt_in(self, aid):
        return self.do_axfer(self.address, aid, Int(0))

    @internal(TealType.uint64)
    def do_create_pool_token(self, a, b):
        return Seq(
            una := AssetParam.unitName(a),
            unb := AssetParam.unitName(b),
            Assert(
                una.hasValue(),
                unb.hasValue(),
            ),
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.AssetConfig,
                    TxnField.config_asset_name: Concat(
                        Bytes("DPT-"), una.value(), Bytes("-"), unb.value()
                    ),
                    TxnField.config_asset_unit_name: Bytes("dpt"),
                    TxnField.config_asset_total: self.total_supply,
                    TxnField.config_asset_decimals: Int(3),
                    TxnField.config_asset_manager: self.address,
                    TxnField.config_asset_reserve: self.address,
                    TxnField.fee: Int(0),
                }
            ),
            InnerTxn.created_asset_id(),
        )

    @internal(TealType.uint64)
    def compute_ratio(self):
        return Seq(
            bal_a := AssetHolding.balance(
                self.address,
                self.asset_a,
            ),
            bal_b := AssetHolding.balance(
                self.address,
                self.asset_b,
            ),
            Assert(
                bal_a.hasValue(),
                bal_b.hasValue(),
            ),
            WideRatio([bal_a.value(), self.scale], [bal_b.value()]),
        )


if __name__ == "__main__":
    ConstantProductAMM().dump("artifacts")



from abc import ABC, abstractmethod
from typing import Literal

from beaker import Application, external
from pyteal import (
    Reject,
    abi,
    Expr,
    Seq,
    ScratchVar,
    Int,
    Suffix,
)


def read_next(vaa: Expr, offset: int, t: abi.BaseType) -> tuple[int, Expr]:
    size = t.type_spec().byte_length_static()
    return offset + size, t.decode(vaa, start_index=Int(offset), length=Int(size))


Bytes32 = abi.StaticBytes[Literal[32]]


class ContractTransferVAA:
    def __init__(self):
        #: Version of VAA
        self.version = abi.Uint8()
        #: Which guardian set to be validated against
        self.index = abi.Uint32()
        #: How many signatures
        self.siglen = abi.Uint8()
        #: TS of message
        self.timestamp = abi.Uint32()
        #: Uniquifying
        self.nonce = abi.Uint32()
        #: The Id of the chain where the message originated
        self.chain = abi.Uint16()
        #: The address of the contract that emitted this message on the origin chain
        self.emitter = abi.Address()
        #: Unique integer representing the index, used for dedupe/ordering
        self.sequence = abi.Uint64()

        self.consistency = abi.Uint8()  # ?

        #: Type of message
        self.type = abi.Uint8()
        #: amount of transfer
        self.amount = abi.make(Bytes32)
        #: asset transferred
        self.contract = abi.make(Bytes32)
        #: Id of the chain the token originated
        self.from_chain = abi.Uint16()
        #: Receiver of the token transfer
        self.to_address = abi.Address()
        #: Id of the chain where the token transfer should be redeemed
        self.to_chain = abi.Uint16()
        #: Amount to pay relayer
        self.fee = abi.make(Bytes32)
        #: Address that sent the transfer
        self.from_address = abi.Address()

        #: Arbitrary byte payload
        self.payload = abi.DynamicBytes()

    def decode(self, vaa: Expr) -> Expr:
        offset = 0
        ops: list[Expr] = []

        offset, e = read_next(vaa, offset, self.version)
        ops.append(e)

        offset, e = read_next(vaa, offset, self.index)
        ops.append(e)

        offset, e = read_next(vaa, offset, self.siglen)
        ops.append(e)

        # Increase offset to skip over sigs && digest
        # since these should be checked by the wormhole core contract
        ops.append(
            (digest_vaa := ScratchVar()).store(
                Suffix(vaa, Int(offset) + (self.siglen.get() * Int(66)))
            )
        )

        # Reset the offset now that we have const length elements
        offset = 0
        offset, e = read_next(digest_vaa.load(), offset, self.timestamp)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.nonce)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.chain)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.emitter)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.sequence)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.consistency)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.type)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.amount)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.contract)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.from_chain)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.to_address)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.to_chain)
        ops.append(e)
        offset, e = read_next(digest_vaa.load(), offset, self.from_address)
        ops.append(e)
        # Rest is payload
        ops.append(self.payload.set(Suffix(digest_vaa.load(), Int(offset))))

        return Seq(*ops)


class WormholeTransfer(Application, ABC):
    """Wormhole Payload3 Message handler

    A Message transfer from another chain to Algorand  using the Wormhole protocol
    will cause this contract to have it's `portal_transfer` method called.
    """

    @external
    def portal_transfer(
        self, vaa: abi.DynamicBytes, *, output: abi.DynamicBytes
    ) -> Expr:
        """portal_transfer accepts a VAA containing information about the transfer and the payload.

        Args:
            vaa: VAA encoded dynamic byte array

        Returns:
            Undefined byte array

        To allow a more flexible interface we publicize that we output generic bytes
        """
        return Seq(
            (ctvaa := ContractTransferVAA()).decode(vaa.get()),
            self.handle_transfer(ctvaa, output=output),
        )

    @abstractmethod
    def handle_transfer(
        self, ctvaa: ContractTransferVAA, *, output: abi.DynamicBytes
    ) -> Expr:
        """handle transfer should be overridden with app specific logic
        needs to be done on transfer

        Args:
            ctvaa: The decoded ContractTransferVAA

        Returns:
            app specific byte array
        """
        return Reject()



import pytest
import pyteal as pt
from beaker.state import (
    ApplicationState,
    AccountState,
    ReservedAccountStateValue,
    ReservedApplicationStateValue,
    ApplicationStateValue,
    AccountStateValue,
    _get_default_for_type,
)

options = pt.CompileOptions(mode=pt.Mode.Application, version=6)

ACCOUNT_VAL_TESTS = [
    # key, stacktype, default, val to set, expected error
    (pt.Bytes("k"), pt.TealType.uint64, pt.Int(1), pt.Int(1), None),
    (pt.Bytes("k"), pt.TealType.uint64, None, pt.Int(1), None),
    (pt.Bytes("k"), pt.TealType.uint64, None, pt.Bytes("abc"), pt.TealTypeError),
    (pt.Int(123), pt.TealType.uint64, None, pt.Int(1), pt.TealTypeError),
    (pt.Bytes("k"), pt.TealType.bytes, None, pt.Bytes("abc"), None),
    (pt.Bytes("k"), pt.TealType.bytes, pt.Bytes("abc"), pt.Bytes("def"), None),
    (pt.Bytes("k"), pt.TealType.bytes, pt.Bytes("abc"), pt.Bytes("def"), None),
    (pt.Bytes("k"), pt.TealType.bytes, None, pt.Int(1), pt.TealTypeError),
    (pt.Bytes("k"), pt.TealType.bytes, pt.Int(1), pt.Bytes("abc"), pt.TealTypeError),
    (pt.Int(123), pt.TealType.bytes, None, pt.Bytes("abc"), pt.TealTypeError),
    (None, pt.TealType.bytes, None, pt.Bytes("abc"), pt.TealInputError),
]


RESERVED_ACCOUNT_VALUE_TESTS = [
    # stack type, max keys, key gen, key_seed, val, expected error
    (pt.TealType.uint64, 1, None, pt.Bytes("abc"), pt.Int(1), None),
    (pt.TealType.bytes, 1, None, pt.Bytes("abc"), pt.Bytes("abc"), None),
    (pt.TealType.bytes, 1, None, pt.abi.String(), pt.Bytes("abc"), None),
    (
        pt.TealType.bytes,
        1,
        pt.Subroutine(pt.TealType.bytes)(
            lambda v: pt.Substring(v, pt.Int(0), pt.Int(1))
        ),
        pt.Bytes("abc"),
        pt.Bytes("abc"),
        None,
    ),
    (
        pt.TealType.bytes,
        1,
        pt.Subroutine(pt.TealType.uint64)(lambda v: pt.Int(1)),
        pt.Bytes("abc"),
        pt.Bytes("abc"),
        pt.TealTypeError,
    ),
    (pt.TealType.uint64, 1, None, pt.Bytes("abc"), pt.Bytes("abc"), pt.TealTypeError),
    (pt.TealType.bytes, 1, None, pt.Bytes("abc"), pt.Int(1), pt.TealTypeError),
    (pt.TealType.uint64, 0, None, pt.Bytes("abc"), pt.Int(1), Exception),
    (pt.TealType.uint64, 17, None, pt.Bytes("abc"), pt.Int(1), Exception),
]


@pytest.mark.parametrize("key, stack_type, default, val, error", ACCOUNT_VAL_TESTS)
def test_local_value(key, stack_type, default, val, error):
    if error is not None:
        with pytest.raises(error):
            do_lv_test(key, stack_type, default, val)
    else:
        do_lv_test(key, stack_type, default, val)


def do_lv_test(key, stack_type, default, val):
    lv = AccountStateValue(stack_type=stack_type, key=key, default=default)

    assert lv.__str__() == f"AccountStateValue (Txn Sender) {key}"

    actual = lv.set(val).__teal__(options)
    expected = pt.App.localPut(pt.Txn.sender(), key, val).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.set_default().__teal__(options)
    expected_default = _get_default_for_type(stack_type, default)
    expected = pt.App.localPut(pt.Txn.sender(), key, expected_default).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get().__teal__(options)
    expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv[pt.Txn.accounts[1]].get().__teal__(options)
    expected = pt.App.localGet(pt.Txn.accounts[1], key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.__teal__(options)
    expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get_maybe().__teal__(options)
    expected = pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    default = _get_default_for_type(stack_type=stack_type, default=None)
    actual = lv.get_else(default).__teal__(options)
    expected = pt.If(
        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)).hasValue(),
        v.value(),
        default,
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.get_must().__teal__(options)
    expected = pt.Seq(
        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)),
        pt.Assert(v.hasValue()),
        v.value(),
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.exists().__teal__(options)
    expected = pt.Seq(
        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)), v.hasValue()
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.delete().__teal__(options)
    expected = pt.App.localDel(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.delete().__teal__(options)
    expected = pt.App.localDel(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


@pytest.mark.parametrize(
    "stack_type, max_keys, key_gen, key_seed, val, error", RESERVED_ACCOUNT_VALUE_TESTS
)
def test_reserved_local_value(stack_type, max_keys, key_gen, key_seed, val, error):
    if error is not None:
        with pytest.raises(error):
            do_reserved_lv_test(stack_type, max_keys, key_gen, key_seed, val)
    else:
        do_reserved_lv_test(stack_type, max_keys, key_gen, key_seed, val)


def do_reserved_lv_test(stack_type, max_keys, key_gen, key_seed, val):
    dlv = ReservedAccountStateValue(
        stack_type=stack_type, max_keys=max_keys, key_gen=key_gen
    )

    lv = dlv[key_seed]

    key = key_seed
    if isinstance(key, pt.abi.BaseType):
        key = key.encode()
    if key_gen is not None:
        key = key_gen(key_seed)

    actual = lv.set(val).__teal__(options)
    expected = pt.App.localPut(pt.Txn.sender(), key, val).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.set_default().__teal__(options)
    expected_default = _get_default_for_type(stack_type, None)
    expected = pt.App.localPut(pt.Txn.sender(), key, expected_default).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get().__teal__(options)
    expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.__teal__(options)
    expected = pt.App.localGet(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get_maybe().__teal__(options)
    expected = pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    default = _get_default_for_type(stack_type=stack_type, default=None)
    actual = lv.get_else(default).__teal__(options)
    expected = pt.If(
        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)).hasValue(),
        v.value(),
        default,
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.get_must().__teal__(options)
    expected = pt.Seq(
        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)),
        pt.Assert(v.hasValue()),
        v.value(),
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.exists().__teal__(options)
    expected = pt.Seq(
        (v := pt.App.localGetEx(pt.Txn.sender(), pt.Int(0), key)), v.hasValue()
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.delete().__teal__(options)
    expected = pt.App.localDel(pt.Txn.sender(), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


APPLICATION_VAL_TESTS = [
    # key, stacktype, default, val to set, static, expected error
    (pt.Bytes("k"), pt.TealType.uint64, pt.Int(1), pt.Int(1), False, None),
    (pt.Bytes("k"), pt.TealType.uint64, None, pt.Int(1), False, None),
    (pt.Bytes("k"), pt.TealType.uint64, None, pt.Bytes("abc"), False, pt.TealTypeError),
    (pt.Int(123), pt.TealType.uint64, None, pt.Int(1), False, pt.TealTypeError),
    (pt.Bytes("k"), pt.TealType.bytes, None, pt.Bytes("abc"), False, None),
    (pt.Bytes("k"), pt.TealType.bytes, pt.Bytes("abc"), pt.Bytes("def"), False, None),
    (pt.Bytes("k"), pt.TealType.bytes, pt.Bytes("abc"), pt.Bytes("def"), False, None),
    (pt.Bytes("k"), pt.TealType.bytes, None, pt.Int(1), False, pt.TealTypeError),
    (pt.Int(123), pt.TealType.bytes, None, pt.Bytes("abc"), False, pt.TealTypeError),
    (pt.Bytes("k"), pt.TealType.bytes, None, pt.Bytes("abc"), True, pt.TealInputError),
    (None, pt.TealType.bytes, None, pt.Bytes("abc"), False, pt.TealInputError),
]


@pytest.mark.parametrize(
    "key, stack_type, default, val, static, error", APPLICATION_VAL_TESTS
)
def test_global_value(key, stack_type, default, val, static, error):
    if error is not None:
        with pytest.raises(error):
            do_gv_test(key, stack_type, default, val, static)
    else:
        do_gv_test(key, stack_type, default, val, static)


def do_gv_test(key, stack_type, default, val, static):
    lv = ApplicationStateValue(
        stack_type=stack_type, key=key, default=default, static=static
    )

    assert lv.__str__() == f"ApplicationStateValue {key}"

    actual = lv.set(val).__teal__(options)

    if static:
        expected = pt.Seq(
            v := pt.App.globalGetEx(pt.Int(0), key),
            pt.Assert(pt.Not(v.hasValue())),
            pt.App.globalPut(key, val),
        ).__teal__(options)
    else:
        expected = pt.App.globalPut(key, val).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.set_default().__teal__(options)
    expected_default = _get_default_for_type(stack_type, default)

    if static:
        expected = pt.Seq(
            v := pt.App.globalGetEx(pt.Int(0), key),
            pt.Assert(pt.Not(v.hasValue())),
            pt.App.globalPut(key, expected_default),
        ).__teal__(options)
    else:
        expected = pt.App.globalPut(key, expected_default).__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.get().__teal__(options)
    expected = pt.App.globalGet(key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.__teal__(options)
    expected = pt.App.globalGet(key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get_maybe().__teal__(options)
    expected = pt.App.globalGetEx(pt.Int(0), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    if stack_type == pt.TealType.uint64 and not static:
        actual = lv.increment().__teal__(options)
        expected = pt.App.globalPut(key, pt.App.globalGet(key) + pt.Int(1)).__teal__(
            options
        )
        with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
            assert actual == expected

        actual = lv.decrement().__teal__(options)
        expected = pt.App.globalPut(key, pt.App.globalGet(key) - pt.Int(1)).__teal__(
            options
        )
        with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
            assert actual == expected

    else:
        with pytest.raises(pt.TealInputError):
            lv.increment()

        with pytest.raises(pt.TealInputError):
            lv.decrement()

    default = _get_default_for_type(stack_type=stack_type, default=None)
    actual = lv.get_else(default).__teal__(options)
    expected = pt.If(
        (v := pt.App.globalGetEx(pt.Int(0), key)).hasValue(), v.value(), default
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.get_must().__teal__(options)
    expected = pt.Seq(
        (v := pt.App.globalGetEx(pt.Int(0), key)), pt.Assert(v.hasValue()), v.value()
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.exists().__teal__(options)
    expected = pt.Seq((v := pt.App.globalGetEx(pt.Int(0), key)), v.hasValue()).__teal__(
        options
    )
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.delete().__teal__(options)
    expected = pt.App.globalDel(key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


RESERVED_APPLICATION_VALUE_TESTS = [
    # stack type, max keys, key gen, key_seed, val, expected error
    (pt.TealType.uint64, 1, None, pt.Bytes("abc"), pt.Int(1), None),
    (pt.TealType.bytes, 1, None, pt.Bytes("abc"), pt.Bytes("abc"), None),
    (pt.TealType.bytes, 1, None, pt.abi.String(), pt.Bytes("abc"), None),
    (
        pt.TealType.bytes,
        1,
        pt.Subroutine(pt.TealType.bytes)(
            lambda v: pt.Substring(v, pt.Int(0), pt.Int(1))
        ),
        pt.Bytes("abc"),
        pt.Bytes("abc"),
        None,
    ),
    (
        pt.TealType.bytes,
        1,
        pt.Subroutine(pt.TealType.uint64)(lambda v: pt.Int(1)),
        pt.Bytes("abc"),
        pt.Bytes("abc"),
        pt.TealTypeError,
    ),
    (pt.TealType.uint64, 1, None, pt.Bytes("abc"), pt.Bytes("abc"), pt.TealTypeError),
    (pt.TealType.bytes, 1, None, pt.Bytes("abc"), pt.Int(1), pt.TealTypeError),
    (pt.TealType.uint64, 0, None, pt.Bytes("abc"), pt.Int(1), Exception),
    (pt.TealType.uint64, 65, None, pt.Bytes("abc"), pt.Int(1), Exception),
]


@pytest.mark.parametrize(
    "stack_type, max_keys, key_gen, key_seed, val, error",
    RESERVED_APPLICATION_VALUE_TESTS,
)
def test_reserved_global_value(stack_type, max_keys, key_gen, key_seed, val, error):
    if error is not None:
        with pytest.raises(error):
            do_reserved_gv_test(stack_type, max_keys, key_gen, key_seed, val)
    else:
        do_reserved_gv_test(stack_type, max_keys, key_gen, key_seed, val)


def do_reserved_gv_test(stack_type, max_keys, key_gen, key_seed, val):
    dlv = ReservedApplicationStateValue(
        stack_type=stack_type, max_keys=max_keys, key_gen=key_gen
    )

    lv = dlv[key_seed]

    key = key_seed
    if isinstance(key, pt.abi.BaseType):
        key = key.encode()

    if key_gen is not None:
        key = key_gen(key_seed)

    actual = lv.set(val).__teal__(options)
    expected = pt.App.globalPut(key, val).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get().__teal__(options)
    expected = pt.App.globalGet(key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.__teal__(options)
    expected = pt.App.globalGet(key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    actual = lv.get_maybe().__teal__(options)
    expected = pt.App.globalGetEx(pt.Int(0), key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    if stack_type == pt.TealType.uint64:
        actual = lv.increment().__teal__(options)
        expected = pt.App.globalPut(key, pt.App.globalGet(key) + pt.Int(1)).__teal__(
            options
        )
        with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
            assert actual == expected

        actual = lv.decrement().__teal__(options)
        expected = pt.App.globalPut(key, pt.App.globalGet(key) - pt.Int(1)).__teal__(
            options
        )
        with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
            assert actual == expected

    else:
        with pytest.raises(pt.TealInputError):
            lv.increment()

        with pytest.raises(pt.TealInputError):
            lv.decrement()

    default = _get_default_for_type(stack_type=stack_type, default=None)
    actual = lv.get_else(default).__teal__(options)
    expected = pt.If(
        (v := pt.App.globalGetEx(pt.Int(0), key)).hasValue(), v.value(), default
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.get_must().__teal__(options)
    expected = pt.Seq(
        (v := pt.App.globalGetEx(pt.Int(0), key)), pt.Assert(v.hasValue()), v.value()
    ).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.exists().__teal__(options)
    expected = pt.Seq((v := pt.App.globalGetEx(pt.Int(0), key)), v.hasValue()).__teal__(
        options
    )
    with pt.TealComponent.Context.ignoreExprEquality(), pt.TealComponent.Context.ignoreScratchSlotEquality():
        assert actual == expected

    actual = lv.delete().__teal__(options)
    expected = pt.App.globalDel(key).__teal__(options)
    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_impl_expr():
    asv = AccountStateValue(pt.TealType.uint64)
    assert asv.has_return() is False
    assert asv.type_of() == pt.TealType.uint64

    asv = ApplicationStateValue(pt.TealType.uint64)
    assert asv.has_return() is False
    assert asv.type_of() == pt.TealType.uint64


def test_application_state():
    statevals = {
        "a": ApplicationStateValue(pt.TealType.uint64),
        "b": ApplicationStateValue(pt.TealType.bytes),
    }
    astate = ApplicationState(statevals)

    assert astate.num_byte_slices == 1
    assert astate.num_uints == 1

    statevals["c"] = ReservedApplicationStateValue(pt.TealType.uint64, max_keys=64)
    with pytest.raises(Exception):
        ApplicationState(statevals)


def test_account_state():
    statevals = {
        "a": AccountStateValue(pt.TealType.uint64),
        "b": AccountStateValue(pt.TealType.bytes),
    }
    astate = AccountState(statevals)

    assert astate.num_byte_slices == 1
    assert astate.num_uints == 1

    statevals["c"] = ReservedAccountStateValue(pt.TealType.uint64, max_keys=16)
    with pytest.raises(Exception):
        AccountState(statevals)



from algosdk.v2client.algod import AlgodClient
from algosdk.v2client.indexer import IndexerClient

DEFAULT_ALGOD_ADDRESS = "http://localhost:4001"
DEFAULT_ALGOD_TOKEN = "a" * 64

DEFAULT_INDEXER_ADDRESS = "http://localhost:8980"
DEFAULT_INDEXER_TOKEN = "a" * 64


def get_algod_client(
    address: str = DEFAULT_ALGOD_ADDRESS, token: str = DEFAULT_ALGOD_TOKEN
) -> AlgodClient:
    """creates a new algod client using the default sandbox parameters"""
    return AlgodClient(token, address)


def get_indexer_client(
    address: str = DEFAULT_INDEXER_ADDRESS, token: str = DEFAULT_INDEXER_TOKEN
) -> IndexerClient:
    """creates a new indexer client using the default sandbox parameters"""
    return IndexerClient(token, address)



from base64 import b64decode
import copy
from typing import Any, cast, Optional

from algosdk.account import address_from_private_key
from algosdk.atomic_transaction_composer import (
    TransactionSigner,
    AccountTransactionSigner,
    MultisigTransactionSigner,
    LogicSigTransactionSigner,
    AtomicTransactionComposer,
    ABIResult,
    ABI_RETURN_HASH,
    TransactionWithSigner,
    abi,
)
from algosdk.future import transaction
from algosdk.logic import get_application_address
from algosdk.source_map import SourceMap
from algosdk.v2client.algod import AlgodClient

from beaker.application import Application, get_method_spec
from beaker.consts import num_extra_program_pages
from beaker.decorators import (
    HandlerFunc,
    MethodHints,
    DefaultArgument,
    DefaultArgumentClass,
)
from beaker.client.state_decode import decode_state
from beaker.client.logic_error import LogicException
from beaker.precompile import AppPrecompile, ProgramAssertion


class ApplicationClient:
    def __init__(
        self,
        client: AlgodClient,
        app: Application,
        app_id: int = 0,
        signer: TransactionSigner = None,
        sender: str = None,
        suggested_params: transaction.SuggestedParams = None,
    ):
        self.client = client
        self.app = app
        self.app_id = app_id
        self.app_addr = get_application_address(app_id) if self.app_id != 0 else None

        self.signer = signer
        self.sender = sender
        if signer is not None and sender is None:
            self.sender = self.get_sender(sender, self.signer)

        self.approval_binary: Optional[bytes] = None
        self.approval_src_map: Optional[SourceMap] = None
        self.approval_asserts: Optional[dict[int, ProgramAssertion]] = None

        self.clear_binary: Optional[bytes] = None
        self.clear_src_map: Optional[SourceMap] = None
        self.clear_asserts: Optional[dict[int, ProgramAssertion]] = None

        self.suggested_params = suggested_params

    def compile(
        self, teal: str, source_map: bool = False
    ) -> tuple[bytes, str, SourceMap]:
        result = self.client.compile(teal, source_map=source_map)
        src_map = None
        if source_map:
            src_map = SourceMap(result["sourcemap"])
        return (b64decode(result["result"]), result["hash"], src_map)

    def build(self) -> None:
        """
        Wraps the Application in an AppPrecompile before calling `compile` on the Precompile which
        recursively compiles all the dependencies (depth first). The result is then used
        to assign the approval and clear state program binaries and src maps.
        """
        if self.approval_binary is not None and self.clear_binary is not None:
            return

        compiled_app = AppPrecompile(self.app)
        compiled_app.compile(self.client)

        self.approval_binary = compiled_app.approval._binary
        self.approval_src_map = compiled_app.approval._map
        self.approval_asserts = compiled_app.approval._asserts

        self.clear_binary = compiled_app.clear._binary
        self.clear_src_map = compiled_app.clear._map
        self.clear_asserts = compiled_app.clear._asserts

    def create(
        self,
        sender: str = None,
        signer: TransactionSigner = None,
        args: list[Any] = None,
        suggested_params: transaction.SuggestedParams = None,
        on_complete: transaction.OnComplete = transaction.OnComplete.NoOpOC,
        extra_pages: int = None,
        **kwargs,
    ) -> tuple[int, str, str]:
        """Submits a signed ApplicationCallTransaction with application id == 0 and the schema and source from the Application passed"""

        self.build()
        assert self.clear_binary is not None and self.approval_binary is not None

        if extra_pages is None:
            extra_pages = num_extra_program_pages(
                self.approval_binary, self.clear_binary
            )

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        atc = AtomicTransactionComposer()
        if self.app.on_create is not None:
            self.add_method_call(
                atc,
                self.app.on_create,
                sender=sender,
                suggested_params=sp,
                on_complete=on_complete,
                approval_program=self.approval_binary,
                clear_program=self.clear_binary,
                global_schema=self.app.app_state.schema(),
                local_schema=self.app.acct_state.schema(),
                extra_pages=extra_pages,
                app_args=args,
                **kwargs,
            )
        else:
            atc.add_transaction(
                TransactionWithSigner(
                    txn=transaction.ApplicationCreateTxn(
                        sender=sender,
                        sp=sp,
                        on_complete=on_complete,
                        approval_program=self.approval_binary,
                        clear_program=self.clear_binary,
                        global_schema=self.app.app_state.schema(),
                        local_schema=self.app.acct_state.schema(),
                        extra_pages=extra_pages,
                        app_args=args,
                        **kwargs,
                    ),
                    signer=signer,
                )
            )

        try:
            create_result = atc.execute(self.client, 4)
        except Exception as e:
            if "logic" in str(e):
                raise self.wrap_approval_exception(e)
            else:
                raise e

        create_txid = create_result.tx_ids[0]

        result = self.client.pending_transaction_info(create_txid)
        app_id = result["application-index"]
        app_addr = get_application_address(app_id)

        self.app_id = app_id
        self.app_addr = app_addr

        return app_id, app_addr, create_txid

    def update(
        self,
        sender: str = None,
        signer: TransactionSigner = None,
        args: list[Any] = None,
        suggested_params: transaction.SuggestedParams = None,
        **kwargs,
    ) -> str:

        """Submits a signed ApplicationCallTransaction with OnComplete set to UpdateApplication and source from the Application passed"""
        self.build()

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        atc = AtomicTransactionComposer()
        if self.app.on_update is not None:
            self.add_method_call(
                atc,
                self.app.on_update,
                on_complete=transaction.OnComplete.UpdateApplicationOC,
                sender=sender,
                suggested_params=sp,
                index=self.app_id,
                approval_program=self.approval_binary,
                clear_program=self.clear_binary,
                app_args=args,
                **kwargs,
            )
        else:
            atc.add_transaction(
                TransactionWithSigner(
                    txn=transaction.ApplicationUpdateTxn(
                        sender=sender,
                        sp=sp,
                        index=self.app_id,
                        approval_program=self.approval_binary,
                        clear_program=self.clear_binary,
                        app_args=args,
                        **kwargs,
                    ),
                    signer=signer,
                )
            )

        try:
            update_result = atc.execute(self.client, 4)
        except Exception as e:
            if "logic" in str(e):
                raise self.wrap_approval_exception(e)
            else:
                raise e

        return update_result.tx_ids[0]

    def opt_in(
        self,
        sender: str = None,
        signer: TransactionSigner = None,
        args: list[Any] = None,
        suggested_params: transaction.SuggestedParams = None,
        **kwargs,
    ) -> str:
        """Submits a signed ApplicationCallTransaction with OnComplete set to OptIn"""

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        atc = AtomicTransactionComposer()
        if self.app.on_opt_in is not None:
            self.add_method_call(
                atc,
                self.app.on_opt_in,
                on_complete=transaction.OnComplete.OptInOC,
                sender=sender,
                suggested_params=sp,
                index=self.app_id,
                app_args=args,
                signer=signer,
                **kwargs,
            )
        else:
            atc.add_transaction(
                TransactionWithSigner(
                    txn=transaction.ApplicationOptInTxn(
                        sender=sender,
                        sp=sp,
                        index=self.app_id,
                        app_args=args,
                        **kwargs,
                    ),
                    signer=signer,
                )
            )

        try:
            opt_in_result = atc.execute(self.client, 4)
        except Exception as e:
            if "logic" in str(e):
                raise self.wrap_approval_exception(e)
            else:
                raise e

        return opt_in_result.tx_ids[0]

    def close_out(
        self,
        sender: str = None,
        signer: TransactionSigner = None,
        args: list[Any] = None,
        suggested_params: transaction.SuggestedParams = None,
        **kwargs,
    ) -> str:
        """Submits a signed ApplicationCallTransaction with OnComplete set to CloseOut"""

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        atc = AtomicTransactionComposer()
        if self.app.on_close_out is not None:
            self.add_method_call(
                atc,
                self.app.on_close_out,
                on_complete=transaction.OnComplete.CloseOutOC,
                sender=sender,
                suggested_params=sp,
                index=self.app_id,
                app_args=args,
                signer=signer,
                **kwargs,
            )
        else:
            atc.add_transaction(
                TransactionWithSigner(
                    txn=transaction.ApplicationCloseOutTxn(
                        sender=sender,
                        sp=sp,
                        index=self.app_id,
                        app_args=args,
                        **kwargs,
                    ),
                    signer=signer,
                )
            )

        try:
            close_out_result = atc.execute(self.client, 4)
        except Exception as e:
            if "logic" in str(e):
                raise self.wrap_approval_exception(e)
            else:
                raise e

        return close_out_result.tx_ids[0]

    def clear_state(
        self,
        sender: str = None,
        signer: TransactionSigner = None,
        args: list[Any] = None,
        suggested_params: transaction.SuggestedParams = None,
        **kwargs,
    ) -> str:

        """Submits a signed ApplicationCallTransaction with OnComplete set to ClearState"""

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        atc = AtomicTransactionComposer()
        if self.app.on_clear_state is not None:
            self.add_method_call(
                atc,
                self.app.on_clear_state,
                on_complete=transaction.OnComplete.ClearStateOC,
                sender=sender,
                suggested_params=sp,
                index=self.app_id,
                app_args=args,
                signer=signer,
                **kwargs,
            )
        else:
            atc.add_transaction(
                TransactionWithSigner(
                    txn=transaction.ApplicationClearStateTxn(
                        sender=sender,
                        sp=sp,
                        index=self.app_id,
                        app_args=args,
                        **kwargs,
                    ),
                    signer=signer,
                )
            )

        clear_state_result = atc.execute(self.client, 4)

        return clear_state_result.tx_ids[0]

    def delete(
        self,
        sender: str = None,
        signer: TransactionSigner = None,
        args: list[Any] = None,
        suggested_params: transaction.SuggestedParams = None,
        **kwargs,
    ) -> str:
        """Submits a signed ApplicationCallTransaction with OnComplete set to DeleteApplication"""

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        atc = AtomicTransactionComposer()
        if self.app.on_delete:
            self.add_method_call(
                atc,
                self.app.on_delete,
                on_complete=transaction.OnComplete.DeleteApplicationOC,
                sender=sender,
                sp=sp,
                index=self.app_id,
                app_args=args,
                signer=signer,
                **kwargs,
            )
        else:
            atc.add_transaction(
                TransactionWithSigner(
                    txn=transaction.ApplicationDeleteTxn(
                        sender=sender,
                        sp=sp,
                        index=self.app_id,
                        app_args=args,
                        **kwargs,
                    ),
                    signer=signer,
                )
            )

        try:
            delete_result = atc.execute(self.client, 4)
        except Exception as e:
            if "logic" in str(e):
                raise self.wrap_approval_exception(e)
            else:
                raise e

        return delete_result.tx_ids[0]

    def prepare(
        self, signer: TransactionSigner = None, sender: str = None, **kwargs
    ) -> "ApplicationClient":

        """makes a copy of the current ApplicationClient and the fields passed"""

        ac = copy.copy(self)
        ac.signer = ac.get_signer(signer)
        ac.sender = ac.get_sender(sender, ac.signer)
        ac.__dict__.update(**kwargs)
        return ac

    def call(
        self,
        method: abi.Method | HandlerFunc,
        sender: str = None,
        signer: TransactionSigner = None,
        suggested_params: transaction.SuggestedParams = None,
        on_complete: transaction.OnComplete = transaction.OnComplete.NoOpOC,
        local_schema: transaction.StateSchema = None,
        global_schema: transaction.StateSchema = None,
        approval_program: bytes = None,
        clear_program: bytes = None,
        extra_pages: int = None,
        accounts: list[str] = None,
        foreign_apps: list[int] = None,
        foreign_assets: list[int] = None,
        boxes: list[tuple[int, bytes]] = None,
        note: bytes = None,
        lease: bytes = None,
        rekey_to: str = None,
        atc: AtomicTransactionComposer = None,
        **kwargs,
    ) -> ABIResult:

        """Handles calling the application"""

        if not isinstance(method, abi.Method):
            method = get_method_spec(method)

        hints = self.method_hints(method.name)

        if atc is None:
            atc = AtomicTransactionComposer()

        atc = self.add_method_call(
            atc,
            method,
            sender,
            signer,
            suggested_params=suggested_params,
            on_complete=on_complete,
            local_schema=local_schema,
            global_schema=global_schema,
            approval_program=approval_program,
            clear_program=clear_program,
            extra_pages=extra_pages,
            accounts=accounts,
            foreign_apps=foreign_apps,
            foreign_assets=foreign_assets,
            note=note,
            lease=lease,
            rekey_to=rekey_to,
            boxes=boxes,
            **kwargs,
        )

        # If its a read-only method, use dryrun (TODO: swap with simulate later?)
        if hints.read_only:
            dr_req = transaction.create_dryrun(self.client, atc.gather_signatures())
            dr_result = self.client.dryrun(dr_req)
            method_results = self._parse_result(
                {0: method}, dr_result["txns"], atc.tx_ids
            )
            return method_results.pop()

        try:
            result = atc.execute(self.client, 4)
        except Exception as e:
            if "logic" in str(e):
                raise self.wrap_approval_exception(e)
            else:
                raise e

        return result.abi_results.pop()

    # TEMPORARY, use SDK one when available
    def _parse_result(
        self,
        methods: dict[int, abi.Method],
        txns: list[dict[str, Any]],
        txids: list[str],
    ) -> list[ABIResult]:
        method_results = []
        for i, tx_info in enumerate(txns):

            raw_value = None
            return_value = None
            decode_error = None

            if i not in methods:
                continue

            # Parse log for ABI method return value
            try:
                if methods[i].returns.type == abi.Returns.VOID:
                    method_results.append(
                        ABIResult(
                            tx_id=txids[i],
                            raw_value=raw_value,
                            return_value=return_value,
                            decode_error=decode_error,
                            tx_info=tx_info,
                            method=methods[i],
                        )
                    )
                    continue

                logs = tx_info["logs"] if "logs" in tx_info else []

                # Look for the last returned value in the log
                if not logs:
                    raise Exception("No logs")

                result = logs[-1]
                # Check that the first four bytes is the hash of "return"
                result_bytes = b64decode(result)
                if len(result_bytes) < 4 or result_bytes[:4] != ABI_RETURN_HASH:
                    raise Exception("no logs")

                raw_value = result_bytes[4:]
                return_value = methods[i].returns.type.decode(raw_value)
            except Exception as e:
                decode_error = e

            method_results.append(
                ABIResult(
                    tx_id=txids[i],
                    raw_value=raw_value,
                    return_value=return_value,
                    decode_error=decode_error,
                    tx_info=tx_info,
                    method=methods[i],
                )
            )

        return method_results

    def add_method_call(
        self,
        atc: AtomicTransactionComposer,
        method: abi.Method | HandlerFunc,
        sender: str = None,
        signer: TransactionSigner = None,
        suggested_params: transaction.SuggestedParams = None,
        on_complete: transaction.OnComplete = transaction.OnComplete.NoOpOC,
        local_schema: transaction.StateSchema = None,
        global_schema: transaction.StateSchema = None,
        approval_program: bytes = None,
        clear_program: bytes = None,
        extra_pages: int = None,
        accounts: list[str] = None,
        foreign_apps: list[int] = None,
        foreign_assets: list[int] = None,
        boxes: list[tuple[int, bytes]] = None,
        note: bytes = None,
        lease: bytes = None,
        rekey_to: str = None,
        **kwargs,
    ):

        """Adds a transaction to the AtomicTransactionComposer passed"""

        sp = self.get_suggested_params(suggested_params)
        signer = self.get_signer(signer)
        sender = self.get_sender(sender, signer)

        if not isinstance(method, abi.Method):
            method = get_method_spec(method)

        hints = self.method_hints(method.name)

        args = []
        for method_arg in method.args:
            name = method_arg.name
            if name in kwargs:
                argument = kwargs[name]
                if type(argument) is dict:
                    if hints.structs is None or name not in hints.structs:
                        raise Exception(f"Name {name} name in struct hints")

                    elems: list[tuple[str, str]] = cast(
                        list[tuple[str, str]], hints.structs[name]["elements"]
                    )

                    argument = [
                        argument[field_name] for field_name, field_type in elems
                    ]

                args.append(argument)

            elif (
                hints.default_arguments is not None and name in hints.default_arguments
            ):
                default_arg = hints.default_arguments[name]
                if default_arg is not None:
                    args.append(self.resolve(default_arg))
            else:
                raise Exception(f"Unspecified argument: {name}")

        atc.add_method_call(
            self.app_id,
            method,
            sender,
            sp,
            signer,
            method_args=args,
            on_complete=on_complete,
            local_schema=local_schema,
            global_schema=global_schema,
            approval_program=approval_program,
            clear_program=clear_program,
            extra_pages=extra_pages,
            accounts=accounts,
            foreign_apps=foreign_apps,
            foreign_assets=foreign_assets,
            boxes=boxes,
            note=note,
            lease=lease,
            rekey_to=rekey_to,
        )

        return atc

    def add_transaction(
        self, atc: AtomicTransactionComposer, txn: transaction.Transaction
    ) -> AtomicTransactionComposer:
        atc.add_transaction(TransactionWithSigner(txn=txn, signer=self.signer))
        return atc

    def fund(self, amt: int, addr: str = None) -> str:
        """convenience method to pay the address passed, defaults to paying the app address for this client from the current signer"""
        sender = self.get_sender()
        signer = self.get_signer()

        sp = self.client.suggested_params()

        rcv = self.app_addr if addr is None else addr

        atc = AtomicTransactionComposer()
        atc.add_transaction(
            TransactionWithSigner(
                txn=transaction.PaymentTxn(sender, sp, rcv, amt),
                signer=signer,
            )
        )
        atc.execute(self.client, 4)
        return atc.tx_ids.pop()

    def get_application_state(self, raw=False) -> dict[bytes | str, bytes | str | int]:
        """gets the global state info for the app id set"""
        app_state = self.client.application_info(self.app_id)
        if "params" not in app_state or "global-state" not in app_state["params"]:
            return {}
        return decode_state(app_state["params"]["global-state"], raw=raw)

    def get_account_state(
        self, account: str = None, raw: bool = False
    ) -> dict[str | bytes, bytes | str | int]:

        """gets the local state info for the app id set and the account specified"""

        if account is None:
            account = self.get_sender()

        acct_state = self.client.account_application_info(account, self.app_id)
        if (
            "app-local-state" not in acct_state
            or "key-value" not in acct_state["app-local-state"]
        ):
            return {}

        return decode_state(acct_state["app-local-state"]["key-value"], raw=raw)

    def get_application_account_info(self) -> dict[str, Any]:
        """gets the account info for the application account"""
        app_state = self.client.account_info(self.app_addr)
        return app_state

    def get_box_names(self) -> list[bytes]:
        box_resp = self.client.application_boxes(self.app_id)
        return [b64decode(box["name"]) for box in box_resp["boxes"]]

    def get_box_contents(self, name: bytes) -> bytes:
        contents = self.client.application_box_by_name(self.app_id, name)
        return b64decode(contents["value"])

    def resolve(self, to_resolve: DefaultArgument) -> Any:
        if to_resolve.resolvable_class == DefaultArgumentClass.Constant:
            return to_resolve.resolve_hint()
        elif to_resolve.resolvable_class == DefaultArgumentClass.GlobalState:
            key = to_resolve.resolve_hint()
            app_state = self.get_application_state(raw=True)
            return app_state[key.encode()]
        elif to_resolve.resolvable_class == DefaultArgumentClass.LocalState:
            key = to_resolve.resolve_hint()
            acct_state = self.get_account_state(self.get_sender(), raw=True)
            return acct_state[key.encode()]
        elif to_resolve.resolvable_class == DefaultArgumentClass.ABIMethod:
            method = abi.Method.undictify(to_resolve.resolve_hint())
            result = self.call(method)
            return result.return_value
        else:
            raise Exception(f"Unrecognized resolver: {to_resolve}")

    def method_hints(self, method_name: str) -> MethodHints:
        if method_name not in self.app.hints:
            return MethodHints()
        return self.app.hints[method_name]

    def get_suggested_params(
        self,
        sp: transaction.SuggestedParams = None,
    ) -> transaction.SuggestedParams:

        if sp is not None:
            return sp

        if self.suggested_params is not None:
            return self.suggested_params

        return self.client.suggested_params()

    def wrap_approval_exception(self, e: Exception) -> Exception:
        if self.app.approval_program is None or self.approval_src_map is None:
            return e

        return LogicException(e, self.app.approval_program, self.approval_src_map)

    def get_signer(self, signer: TransactionSigner = None) -> TransactionSigner:
        if signer is not None:
            return signer

        if self.signer is not None:
            return self.signer

        raise Exception("No signer provided")

    def get_sender(self, sender: str = None, signer: TransactionSigner = None) -> str:
        if sender is not None:
            return sender

        if signer is None and self.sender is not None:
            return self.sender

        signer = self.get_signer(signer)

        match signer:
            case AccountTransactionSigner():  # type: ignore
                return address_from_private_key(signer.private_key)
            case MultisigTransactionSigner():  # type: ignore
                return signer.msig.address()
            case LogicSigTransactionSigner():  # type: ignore
                return signer.lsig.address()

        raise Exception("No sender provided")



from typing import Literal
from pyteal import *
from beaker import *

from beaker.application import get_method_signature
from beaker.lib.storage import Mapping, List

# Use a box per member to denote membership parameters
class MembershipRecord(abi.NamedTuple):
    role: abi.Field[abi.Uint8]
    voted: abi.Field[abi.Bool]


Affirmation = abi.StaticBytes[Literal[64]]

BoxFlatMinBalance = 2500
BoxByteMinBalance = 400
AssetMinBalance = 100000


def axfer(
    asset_id: Expr,
    amount: Expr,
    receiver: Expr,
    extra: dict[TxnField, Expr] | None = None,
) -> dict[TxnField, Expr]:
    extra = extra if extra is not None else {}
    return {
        TxnField.type_enum: TxnType.AssetTransfer,
        TxnField.xfer_asset: asset_id,
        TxnField.asset_amount: amount,
        TxnField.asset_receiver: receiver,
    } | extra


def clawback_axfer(
    asset_id: Expr,
    amount: Expr,
    receiver: Expr,
    clawback_addr: Expr,
    extra: dict[TxnField, Expr] | None = None,
) -> dict[TxnField, Expr]:
    return axfer(asset_id, amount, receiver, extra) | {
        TxnField.asset_sender: clawback_addr,
    }


class MembershipClub(Application):
    ####
    # Box abstractions

    # A Mapping will create a new box for every unique key, taking a data type for key and value
    # Only static types can provide information about the max size (and min balance required).
    membership_records = Mapping(abi.Address, MembershipRecord)

    # A Listing is a simple list, initialized with some _static_ data type and a length
    affirmations = List(Affirmation, 10)

    #####
    # State vars

    membership_token = ApplicationStateValue(
        TealType.uint64,
        static=True,
        descr="The asset that represents membership of this club",
    )

    #####
    # Math for determining min balance based on expected size of boxes
    _max_members = 1000
    _member_box_size = abi.size_of(MembershipRecord)
    _min_balance = (
        AssetMinBalance  # Cover min bal for member token
        + (BoxFlatMinBalance + (_member_box_size * BoxByteMinBalance))
        * _max_members  # cover min bal for member record boxes we might create
        + (
            BoxFlatMinBalance + (affirmations._box_size * BoxByteMinBalance)
        )  # cover min bal for affirmation box
    )
    ####

    MaxMembers = Int(_max_members)
    MinimumBalance = Int(_min_balance)

    @external(authorize=Authorize.only(Global.creator_address()))
    def bootstrap(
        self,
        seed: abi.PaymentTransaction,
        token_name: abi.String,
        *,
        output: abi.Uint64,
    ):
        """create membership token and receive initial seed payment"""
        return Seq(
            Assert(
                seed.get().receiver() == self.address,
                comment="payment must be to app address",
            ),
            Assert(
                seed.get().amount() >= self.MinimumBalance,
                comment=f"payment must be for >= {self._min_balance}",
            ),
            Pop(self.affirmations.create()),
            InnerTxnBuilder.Execute(
                {
                    TxnField.type_enum: TxnType.AssetConfig,
                    TxnField.config_asset_name: token_name.get(),
                    TxnField.config_asset_total: self.MaxMembers,
                    TxnField.config_asset_default_frozen: Int(1),
                    TxnField.config_asset_manager: self.address,
                    TxnField.config_asset_clawback: self.address,
                    TxnField.config_asset_freeze: self.address,
                    TxnField.config_asset_reserve: self.address,
                    TxnField.fee: Int(0),
                }
            ),
            self.membership_token.set(InnerTxn.created_asset_id()),
            output.set(self.membership_token),
        )

    @external(authorize=Authorize.only(Global.creator_address()))
    def remove_member(self, member: abi.Address):
        return Pop(self.membership_records[member].delete())

    @external(authorize=Authorize.only(Global.creator_address()))
    def add_member(
        self, new_member: abi.Account, membership_token: abi.Asset = membership_token
    ):
        return Seq(
            (role := abi.Uint8()).set(Int(0)),
            (voted := abi.Bool()).set(consts.FALSE),
            (mr := MembershipRecord()).set(role, voted),
            self.membership_records[new_member.address()].set(mr),
            InnerTxnBuilder.Execute(
                clawback_axfer(
                    self.membership_token,
                    Int(1),
                    new_member.address(),
                    self.address,
                    extra={TxnField.fee: Int(0)},
                )
            ),
        )

    @external(read_only=True)
    def get_membership_record(self, member: abi.Address, *, output: MembershipRecord):
        return self.membership_records[member].store_into(output)

    @external(authorize=Authorize.holds_token(membership_token))
    def set_affirmation(
        self,
        idx: abi.Uint16,
        affirmation: Affirmation,
        membership_token: abi.Asset = membership_token,
    ):
        return self.affirmations[idx.get()].set(affirmation)

    @external(authorize=Authorize.holds_token(membership_token))
    def get_affirmation(
        self, membership_token: abi.Asset = membership_token, *, output: Affirmation
    ):
        return output.set(
            self.affirmations[Global.round() % self.affirmations.elements]
        )


class AppMember(Application):

    membership_token = ApplicationStateValue(TealType.uint64)
    club_app_id = ApplicationStateValue(TealType.uint64)
    last_affirmation = ApplicationStateValue(TealType.bytes)

    @external(authorize=Authorize.only(Global.creator_address()))
    def bootstrap(
        self,
        seed: abi.PaymentTransaction,
        app_id: abi.Application,
        membership_token: abi.Asset,
    ):
        return Seq(
            # Set app id
            self.club_app_id.set(app_id.application_id()),
            # Set membership token
            self.membership_token.set(membership_token.asset_id()),
            # Opt in to membership token
            InnerTxnBuilder.Execute(
                axfer(membership_token.asset_id(), Int(0), self.address)
                | {TxnField.fee: Int(0)},
            ),
        )

    @external
    def get_affirmation(
        self,
        member_token: abi.Asset = membership_token,
        club_app: abi.Application = club_app_id,
    ):
        return Seq(
            InnerTxnBuilder.ExecuteMethodCall(
                app_id=self.club_app_id,
                method_signature=get_method_signature(MembershipClub.get_affirmation),
                args=[member_token],
            ),
            self.last_affirmation.set(Suffix(InnerTxn.last_log(), Int(4))),
        )


if __name__ == "__main__":
    MembershipClub().dump("./artifacts")



from pyteal import (
    abi,
    Int,
    BoxCreate,
    BoxExtract,
    Expr,
    BoxReplace,
    Bytes,
    TealType,
    TealTypeError,
)


class List:
    """List stores a list of static types in a box, named as the class attribute unless an overriding name is provided"""

    def __init__(self, value_type: type[abi.BaseType], elements: int, name: str = None):
        ts = abi.type_spec_from_annotation(value_type)

        assert not ts.is_dynamic(), "Expected static type for value"
        assert (
            ts.byte_length_static() * elements < 32e3
        ), "Cannot be larger than MAX_BOX_SIZE"

        # Will be set later if its part of an Application
        self.name: Expr | None = None
        if name is not None:
            self.name = Bytes(name)

        self.value_type = ts

        self._element_size = ts.byte_length_static()
        self.element_size = Int(self._element_size)

        self._elements = elements
        self.elements = Int(self._elements)

        self._box_size = self._element_size * self._elements
        self.box_size = Int(self._box_size)

    def create(self) -> Expr:
        """creates a box with the given name and with a size that will allow storage of the number of the element specified."""
        assert self.name is not None
        return BoxCreate(self.name, self.box_size)

    def __getitem__(self, idx: Expr) -> "ListElement":
        assert self.name is not None
        return ListElement(self.name, self.element_size, idx)


class ListElement(Expr):
    def __init__(self, name: Expr, element_size: Expr, idx: Expr):
        if name.type_of() != TealType.bytes:
            raise TealTypeError(name.type_of(), TealType.bytes)

        if element_size.type_of() != TealType.uint64:
            raise TealTypeError(element_size.type_of(), TealType.uint64)

        if element_size.type_of() != TealType.uint64:
            raise TealTypeError(idx.type_of(), TealType.uint64)

        self.name = name
        self.element_size = element_size
        self.idx = idx

    def store_into(self, val: abi.BaseType) -> Expr:
        """decode the bytes from this list element into the instance of the type provided

        Args:
            val: An instance of the type to decode into
        """
        return val.decode(self.get())

    def get(self) -> Expr:
        """get the bytes for this element in the list"""
        return BoxExtract(self.name, self.element_size * self.idx, self.element_size)

    def set(self, val: abi.BaseType) -> Expr:
        """set the bytes for this element in the list

        Args:
            The value to write into the list at the given index
        """
        return BoxReplace(self.name, self.element_size * self.idx, val.encode())

    def __str__(self) -> str:
        return f"List Element: {self.name}[{self.idx}]"

    def __teal__(self, compile_options):
        return self.get().__teal__(compile_options)

    def has_return(self):
        return False

    def type_of(self):
        return TealType.bytes



import json
from algosdk.abi import ABIType
from beaker import *

if __name__ == "__main__":
    from contract import OracleDataCache
else:
    from .contract import OracleDataCache


base_vaa = bytes.fromhex(
    "010000000001008049340af360a47103a962108cb57b9deebcc99e8e6ddeca1a"
    + "1fb025413a62ac2cae4abd6b7e0ce7fc5a6bc99536387a3827cbbb0c710c81213"
    + "a417cb59b89de01630d06ae0000000000088edf5b0e108c3a1a0a4b704cc89591"
    + "f2ad8d50df24e991567e640ed720a94be20000000000000004000300000000000"
    + "00000000000000000000000000000000000000000000000000064000000000000"
    + "0000000000000000000000000000000000000000000000000000000f6463fab1a"
    + "45027a3c70781ae588e4e6661d21a7c19535a5d6b4f4c3164a13be1000f0b7ef3"
    + "fcf3f8d9efc458695dc7bd7e534080ac7b48f2b881fd3063b1308f0648"
)

# Get the codec to decode the stored value
oracle_data_codec = ABIType.from_string(str(OracleDataCache.OracleData().type_spec()))


def demo():
    app_client = client.ApplicationClient(
        sandbox.get_algod_client(),
        OracleDataCache(),
        signer=sandbox.get_accounts().pop().signer,
    )

    # Deploy the app on chain
    app_client.create()

    # Make up some fake oracle data and send it to the contract
    base_ts = 1661802300
    base_price = 10000
    for x in range(10):
        fauxracle_data = {
            "ts": base_ts + x * 60,
            "price": base_price + x,
            "confidence": 9999,
        }

        app_client.call(
            OracleDataCache.portal_transfer,
            vaa=base_vaa + json.dumps(fauxracle_data).encode(),
        )

    # Get the current app state
    app_state = app_client.get_application_state(raw=True)
    for v in app_state.values():
        ts, price, confidence = oracle_data_codec.decode(v)
        print(f"ts: {ts}, price: {price}, confidence: {confidence}")


if __name__ == "__main__":
    demo()



from enum import Enum
from algosdk.v2client.algod import AlgodClient
from algosdk.v2client.indexer import IndexerClient


class Network(Enum):
    """Provides consistent way to reference the most common network options"""

    MainNet = "MainNet"
    TestNet = "TestNet"
    BetaNet = "BetaNet"
    SandNet = "SandNet"


class APIProvider:
    """abstract class to provide interface for API providers"""

    algod_hosts: dict[Network, str] = {}
    indexer_hosts: dict[Network, str] = {}

    def __init__(self, network: Network):
        self.network = network

    def algod(self, token: str = "") -> AlgodClient:
        """return an algod client based on the provider used and network it was initialized with"""
        if self.network not in self.algod_hosts:
            raise Exception(f"Unrecognized network: {self.network}")
        return AlgodClient(token, self.algod_hosts[self.network])

    def indexer(self, token: str = "") -> IndexerClient:
        """return an indexer client based on the provider used and network it was initialized with"""
        if self.network not in self.indexer_hosts:
            raise Exception(f"Unrecognized network: {self.network}")
        return IndexerClient(token, self.indexer_hosts[self.network])


class AlgoNode(APIProvider):
    algod_hosts = {
        Network.MainNet: "https://mainnet-api.algonode.cloud",
        Network.TestNet: "https://testnet-api.algonode.cloud",
        Network.BetaNet: "https://betanet-api.algonode.cloud",
    }

    indexer_hosts = {
        Network.MainNet: "https://mainnet-idx.algonode.cloud",
        Network.TestNet: "https://testnet-idx.algonode.cloud",
        Network.BetaNet: "https://betanet-idx.algonode.cloud",
    }


class AlgoExplorer(APIProvider):
    algod_hosts = {
        Network.MainNet: "https://node.algoexplorerapi.io",
        Network.TestNet: "https://node.testnet.algoexplorerapi.io",
        Network.BetaNet: "https://node.betanet.algoexplorerapi.io",
    }

    indexer_hosts = {
        Network.MainNet: "https://algoindexer.algoexplorerapi.io",
        Network.TestNet: "https://algoindexer.testnet.algoexplorerapi.io",
        Network.BetaNet: "https://algoindexer.betanet.algoexplorerapi.io",
    }


class PureStake(APIProvider):
    algod_hosts = {
        Network.MainNet: "https://mainnet-algorand.api.purestake.io/ps1",
        Network.TestNet: "https://testnet-algorand.api.purestake.io/ps1",
        Network.BetaNet: "https://betanet-algorand.api.purestake.io/ps1",
    }

    indexer_hosts = {
        Network.MainNet: "https://mainnet-algorand.api.purestake.io/idx2",
        Network.TestNet: "https://testnet-algorand.api.purestake.io/idx2",
        Network.BetaNet: "https://betanet-algorand.api.purestake.io/idx2",
    }

    token_header = "X-API-Key"

    # Purestake requires a different header, so we override the
    # existing methods but re-use them to generate the client
    def algod(self, token: str = "") -> AlgodClient:
        algod_client = super().algod()
        algod_client.headers = {self.token_header: token}

    def indexer(self, token: str = "") -> IndexerClient:
        indexer_client = super().algod()
        indexer_client.headers = {self.token_header: token}


class Sandbox(APIProvider):
    default_host = "http://localhost"
    default_algod_port: int = 4001
    default_indexer_port: int = 8980
    default_token: str = "a" * 64

    # purposely doesn't check which network because it may
    # be set up to use mainnet/testnet
    def algod(self, token: str = default_token) -> AlgodClient:
        address = f"{self.default_host}:{self.default_algod_port}"
        return AlgodClient(token, address)

    def indexer(self, token: str = default_token) -> IndexerClient:
        address = f"{self.default_host}:{self.default_indexer_port}"
        return IndexerClient(token, address)



from typing import Any
from base64 import b64decode


def str_or_hex(v: bytes) -> str:
    decoded: str = ""
    try:
        decoded = v.decode("utf-8")
    except Exception:
        decoded = v.hex()

    return decoded


def decode_state(
    state: list[dict[str, Any]], raw=False
) -> dict[str | bytes, bytes | str | int]:

    decoded_state: dict[str | bytes, bytes | str | int] = {}

    for sv in state:

        raw_key = b64decode(sv["key"])

        key: str | bytes = raw_key if raw else str_or_hex(raw_key)
        val: str | bytes | int

        match sv["value"]["type"]:
            case 1:
                raw_val = b64decode(sv["value"]["bytes"])
                val = raw_val if raw else str_or_hex(raw_val)
            case 2:
                val = sv["value"]["uint"]

        decoded_state[key] = val
    return decoded_state



from pyteal import (
    Assert,
    BitLen,
    Btoi,
    Bytes,
    BytesDiv,
    BytesGt,
    BytesMod,
    Concat,
    Extract,
    GetByte,
    If,
    Int,
    Itob,
    Len,
    ScratchVar,
    Seq,
    Subroutine,
    Substring,
    TealType,
)

from ..math import pow10

# Magic number to convert between ascii chars and integers
_ascii_zero = 48
_ascii_nine = _ascii_zero + 9
ascii_zero = Int(_ascii_zero)
ascii_nine = Int(_ascii_nine)


@Subroutine(TealType.uint64)
def ascii_to_int(arg):
    """ascii_to_int converts the integer representing a character in ascii to the actual integer it represents

    Args:
        arg: uint64 in the range 48-57 that is to be converted to an integer

    Returns:
        uint64 that is the value the ascii character passed in represents

    """
    return Seq(Assert(arg >= ascii_zero), Assert(arg <= ascii_nine), arg - ascii_zero)


@Subroutine(TealType.bytes)
def int_to_ascii(arg):
    """int_to_ascii converts an integer to the ascii byte that represents it"""
    return Extract(Bytes("0123456789"), arg, Int(1))


@Subroutine(TealType.uint64)
def atoi(a):
    """atoi converts a byte string representing a number to the integer value it represents"""
    return If(
        Len(a) > Int(0),
        (ascii_to_int(GetByte(a, Int(0))) * pow10(Len(a) - Int(1)))
        + atoi(Substring(a, Int(1), Len(a))),
        Int(0),
    )


@Subroutine(TealType.bytes)
def itoa(i):
    """itoa converts an integer to the ascii byte string it represents"""
    return If(
        i == Int(0),
        Bytes("0"),
        Concat(
            If(i / Int(10) > Int(0), itoa(i / Int(10)), Bytes("")),
            int_to_ascii(i % Int(10)),
        ),
    )


@Subroutine(TealType.bytes)
def witoa(i):
    """witoa converts an byte string interpreted as an integer to the ascii byte string it represents"""
    return If(
        BitLen(i) == Int(0),
        Bytes("0"),
        Concat(
            If(
                BytesGt(BytesDiv(i, Bytes("base16", "A0")), Bytes("base16", "A0")),
                witoa(BytesDiv(i, Bytes("base16", "A0"))),
                Bytes(""),
            ),
            int_to_ascii(Btoi(BytesMod(i, Bytes("base16", "A0")))),
        ),
    )


@Subroutine(TealType.bytes)
def head(s):
    """head gets the first byte from a bytestring, returns as bytes"""
    return Extract(s, Int(0), Int(1))


@Subroutine(TealType.bytes)
def tail(s):
    """tail returns the string with the first character removed"""
    return Substring(s, Int(1), Len(s))


@Subroutine(TealType.bytes)
def suffix(s, n):
    """suffix returns the last n bytes of a given byte string"""
    return Substring(s, Len(s) - n, Len(s))


@Subroutine(TealType.bytes)
def prefix(s, n):
    """prefix returns the first n bytes of a given byte string"""
    return Substring(s, Int(0), n)


@Subroutine(TealType.bytes)
def rest(s, n):
    """prefix returns the first n bytes of a given byte string"""
    return Substring(s, n, Len(s))


def encode_uvarint(val):
    """
    Returns the uvarint encoding of an integer

    Useful in the case that the bytecode for a contract is being populated, since
    integers in a contract are uvarint encoded

    This subroutine is recursive, the first call should include
    the integer to be encoded and an empty bytestring

    """

    @Subroutine(TealType.bytes)
    def encode_uvarint_impl(val, b):
        buff = ScratchVar()
        return Seq(
            buff.store(b),
            Concat(
                buff.load(),
                If(
                    val >= Int(128),
                    encode_uvarint_impl(
                        val >> Int(7),
                        Extract(Itob((val & Int(255)) | Int(128)), Int(7), Int(1)),
                    ),
                    Extract(Itob(val & Int(255)), Int(7), Int(1)),
                ),
            ),
        )

    return encode_uvarint_impl(val, Bytes(""))



import random
import string
from typing import cast
import algosdk.future.transaction as txns
from algosdk.atomic_transaction_composer import *
from pyteal import *
from beaker import *
from beaker.precompile import LSigPrecompile

# Simple logic sig, will approve _any_ transaction
# Used to expand our apps available state by
# creating unique account that will do whatever we need.
# In this case, we need it to opt in and rekey to the app address
class KeySig(LogicSignature):
    nonce = TemplateVariable(TealType.bytes)

    def evaluate(self):
        return Approve()


# App that needs lots of storage so we use the local storage of
# unique lsig accounts that have been rekeyed to the app address.
# This allows us to use the local storage of the unique accounts
# to get an extra 2k of storage for each account
class DiskHungry(Application):
    # Reserve all 16 keys for the blob in local state
    data = AccountStateBlob(keys=16)

    # Signal to beaker that this should be compiled
    # prior to compiling the main application
    tmpl_acct = LSigPrecompile(KeySig(version=6))

    # Add account during opt in  by checking the sender against the address
    # we expect given the precompile && nonce
    @opt_in
    def add_account(self, nonce: abi.DynamicBytes):
        return Seq(
            Assert(
                # Make sure the opt-in'er is our lsig
                Txn.sender() == self.tmpl_acct.logic.template_hash(nonce.get()),
                # and that its being rekeyed to us
                Txn.rekey_to() == self.address,
            ),
            self.initialize_account_state(),
        )

    # Inline these
    def byte_idx(self, bit_idx) -> Int:
        return bit_idx / Int(8)

    def bit_in_byte_idx(self, bit_idx) -> Int:
        return bit_idx % Int(8)

    @external
    def flip_bit(self, nonce_acct: abi.Account, bit_idx: abi.Uint32):
        """
        Allows caller to flip a bit at a given index for some account that has already opted in
        """

        return Seq(
            # Read byte
            (byte := ScratchVar()).store(
                self.data[nonce_acct.address()].read_byte(self.byte_idx(bit_idx.get()))
            ),
            # Flip bit
            byte.store(
                SetBit(
                    byte.load(),
                    self.bit_in_byte_idx(bit_idx.get()),
                    Not(GetBit(byte.load(), self.bit_in_byte_idx(bit_idx.get()))),
                )
            ),
            # Write byte
            self.data[nonce_acct.address()].write_byte(
                self.byte_idx(bit_idx.get()), byte.load()
            ),
        )


def demo():

    # Instantiate our app, since we're using a precompile we want this instance
    # to access the precompiled lsig
    app = DiskHungry()

    # Create app client
    app_client = client.ApplicationClient(
        client=sandbox.get_algod_client(),
        app=app,
        signer=sandbox.get_accounts().pop().signer,
    )

    # Deploy the app
    app_client.create()

    # Create 10 random nonces for unique lsig accounts
    # and make them opt in to the app
    for _ in range(10):
        # Populate the binary template with the random nonce and get back
        # a Signer obj to submit transactions
        nonce: str = get_nonce()
        lsig_signer: LogicSigTransactionSigner = app.tmpl_acct.template_signer(nonce)

        print(
            f"Creating templated lsig with nonce {nonce} and address {lsig_signer.lsig.address()}"
        )

        # Create the account and opt it into the app, also rekeys it to the app address
        create_and_opt_in_account(
            app_client, app_client.prepare(signer=lsig_signer), nonce
        )

        # Max is 8 (bits per byte) * 127 (bytes per key) * 16 (max keys) == 16256
        idx: int = 16255
        app_client.call(
            DiskHungry.flip_bit, nonce_acct=lsig_signer.lsig.address(), bit_idx=idx
        )

        # Get the full state for the lsig we used to store this bit
        acct_state: dict[bytes, bytes] = app_client.get_account_state(
            lsig_signer.lsig.address(), raw=True
        )

        # Make sure the blob is in the right order
        blob: bytes = b"".join([acct_state[x.to_bytes(1, "big")] for x in range(16)])

        # Did the expected byte have the expected integer value?
        assert int(blob[idx // 8]) == 2 ** (idx % 8)
        print(f"bit set correctly at index {idx}")


def get_nonce(n: int = 10) -> bytes:
    return ("".join(random.choice(string.ascii_uppercase) for _ in range(n))).encode()


def create_and_opt_in_account(
    user_client: client.ApplicationClient,
    lsig_client: client.ApplicationClient,
    nonce: bytes,
):
    sp = user_client.get_suggested_params()
    lsig_address = cast(LogicSigTransactionSigner, lsig_client.signer).lsig.address()

    atc = AtomicTransactionComposer()

    # Give the lsig 2 algo for min balance (More than needed)
    user_client.add_transaction(
        atc,
        txns.PaymentTxn(
            sender=user_client.sender,
            sp=sp,
            receiver=lsig_address,
            amt=2 * consts.algo,
        ),
    )
    # Add opt in method call on behalf of lsig
    lsig_client.add_method_call(
        atc,
        DiskHungry.add_account,
        suggested_params=sp,
        nonce=nonce,
        rekey_to=lsig_client.app_addr,
        on_complete=txns.OnComplete.OptInOC,
    )

    # Run it
    atc.execute(user_client.client, 4)


if __name__ == "__main__":
    demo()
